# This file is part of django-ca (https://github.com/mathiasertl/django-ca).
#
# django-ca is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# django-ca is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with django-ca.  If not,
# see <http://www.gnu.org/licenses/>.

"""Central functions to load CA key and cert as PKey/X509 objects."""

import binascii
import os
import re
import shlex
from datetime import datetime
from datetime import timedelta
from ipaddress import ip_address
from ipaddress import ip_network
from urllib.parse import urlparse

import idna

from asn1crypto.core import OctetString
from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import dsa
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives.serialization import Encoding
from cryptography.x509.oid import NameOID

from django.core.files.storage import get_storage_class
from django.core.serializers.json import DjangoJSONEncoder
from django.core.validators import URLValidator
from django.utils.encoding import force_bytes
from django.utils.encoding import force_str
from django.utils.functional import Promise
from django.utils.translation import gettext_lazy as _

from . import ca_settings

# List of possible subject fields, in order
SUBJECT_FIELDS = ['C', 'ST', 'L', 'O', 'OU', 'CN', 'emailAddress', ]

# Description strings for various X509 extensions, taken from "man x509v3_config".
EXTENDED_KEY_USAGE_DESC = _('Purposes for which the certificate public key can be used for.')
KEY_USAGE_DESC = _('Permitted key usages.')

#: Regular expression to match RDNs out of a full x509 name.
NAME_RE = re.compile(r'(?:/+|\A)\s*(?P<field>[^\s]*?)\s*'
                     r'=(?P<quote>[\'"])?\s*(?P<content>(?(quote).*?|[^/]*))\s*'
                     r'(?(quote)(?<!\\)(?P=quote))', re.I)

#: Regular expression to match general names.
GENERAL_NAME_RE = re.compile('^(email|URI|IP|DNS|RID|dirName|otherName):(.*)', flags=re.I)

#: Regular expression matching hexlified certificate serials
SERIAL_RE = re.compile('^([0-9A-F][0-9A-F]:?)+[0-9A-F][0-9A-F]?$')

SAN_NAME_MAPPINGS = {
    x509.DNSName: 'DNS',
    x509.RFC822Name: 'email',
    x509.DirectoryName: 'dirname',
    x509.UniformResourceIdentifier: 'URI',
    x509.IPAddress: 'IP',
    x509.RegisteredID: 'RID',
    x509.OtherName: 'otherName',
}

#: Map OID objects to IDs used in subject strings
OID_NAME_MAPPINGS = {
    NameOID.COUNTRY_NAME: 'C',
    NameOID.STATE_OR_PROVINCE_NAME: 'ST',
    NameOID.LOCALITY_NAME: 'L',
    NameOID.ORGANIZATION_NAME: 'O',
    NameOID.ORGANIZATIONAL_UNIT_NAME: 'OU',
    NameOID.COMMON_NAME: 'CN',
    NameOID.EMAIL_ADDRESS: 'emailAddress',
    NameOID.SERIAL_NUMBER: 'serialNumber',
    NameOID.JURISDICTION_COUNTRY_NAME: 'jurisdictionCountryName',
    NameOID.JURISDICTION_STATE_OR_PROVINCE_NAME: 'jurisdictionStateOrProvinceName',
    NameOID.BUSINESS_CATEGORY: "businessCategory",
    NameOID.POSTAL_CODE: "postalCode",
    NameOID.STREET_ADDRESS: "streetAddress",
}

# same, but reversed
NAME_OID_MAPPINGS = {v: k for k, v in OID_NAME_MAPPINGS.items()}

# Some OIDs can occur multiple times
MULTIPLE_OIDS = (
    NameOID.ORGANIZATIONAL_UNIT_NAME,
    NameOID.STREET_ADDRESS,
)

# uppercase values as keys for normalizing case
NAME_CASE_MAPPINGS = {v.upper(): v for v in OID_NAME_MAPPINGS.values()}


class LazyEncoder(DjangoJSONEncoder):
    """Encoder that also encodes strings translated with gettext_lazy."""

    def default(self, o):
        if isinstance(o, Promise):
            return force_str(o)
        return super().default(o)


def sort_name(subject):
    """Returns the subject in the correct order for a x509 subject."""
    return sorted(subject, key=lambda e: SUBJECT_FIELDS.index(e[0]))


def encode_url(url):
    """IDNA encoding for domains in URLs.

    Examples::

        >>> encode_url('https://example.com')
        'https://example.com'
        >>> encode_url('https://exämple.com/foobar')
        'https://xn--exmple-cua.com/foobar'
        >>> encode_url('https://exämple.com:8000/foobar')
        'https://xn--exmple-cua.com:8000/foobar'
    """
    parsed = urlparse(url)
    if parsed.port:
        hostname = idna.encode(parsed.hostname).decode('utf-8')
        parsed = parsed._replace(netloc='%s:%s' % (hostname, parsed.port))
    else:
        parsed = parsed._replace(netloc=idna.encode(parsed.netloc).decode('utf-8'))
    return parsed.geturl()


def encode_dns(name):
    """IDNA encoding for domains.

    Examples::

        >>> encode_dns('example.com')
        'example.com'
        >>> encode_dns('exämple.com')
        'xn--exmple-cua.com'
        >>> encode_dns('.exämple.com')
        '.xn--exmple-cua.com'
        >>> encode_dns('*.exämple.com')
        '*.xn--exmple-cua.com'
    """
    if name.startswith('*.'):
        return '*.%s' % idna.encode(name[2:]).decode('utf-8')
    if name.startswith('.'):
        return '.%s' % idna.encode(name[1:]).decode('utf-8')
    return idna.encode(name).decode('utf-8')


def format_name(subject):
    """Convert a subject into the canonical form for distinguished names.

    This function does not take care of sorting the subject in any meaningful order.

    Examples::

        >>> format_name([('CN', 'example.com'), ])
        '/CN=example.com'
        >>> format_name([('CN', 'example.com'), ('O', "My Organization"), ])
        '/CN=example.com/O=My Organization'
    """
    if isinstance(subject, x509.Name):
        subject = [(OID_NAME_MAPPINGS[s.oid], s.value) for s in subject]

    return '/%s' % ('/'.join(['%s=%s' % (force_str(k), force_str(v)) for k, v in subject]))


def format_relative_name(name):
    """Convert a relative name (RDN) into a canonical form.

    Examples::

        >>> format_relative_name([('C', 'AT'), ('CN', 'example.com')])
        '/C=AT/CN=example.com'
        >>> format_relative_name(x509.RelativeDistinguishedName([
        ...     x509.NameAttribute(NameOID.COMMON_NAME, 'example.com')
        ... ]))
        '/CN=example.com'
    """
    if isinstance(name, x509.RelativeDistinguishedName):
        name = [(OID_NAME_MAPPINGS[s.oid], s.value) for s in name]

    return '/%s' % ('/'.join(['%s=%s' % (force_str(k), force_str(v)) for k, v in name]))


def format_general_name(name):
    """Format a single general name.

    >>> import ipaddress
    >>> format_general_name(x509.DNSName('example.com'))
    'DNS:example.com'
    >>> format_general_name(x509.IPAddress(ipaddress.IPv4Address('127.0.0.1')))
    'IP:127.0.0.1'
    """

    if isinstance(name, x509.DirectoryName):
        value = format_name(name.value)
    else:
        value = name.value
    return '%s:%s' % (SAN_NAME_MAPPINGS[type(name)], value)


def is_power2(num):
    """Return True if num is a power of 2.

    >>> is_power2(4)
    True
    >>> is_power2(3)
    False
    """
    return num != 0 and ((num & (num - 1)) == 0)


def multiline_url_validator(value):
    """Validate that a TextField contains one valid URL per line.

    .. seealso:: https://docs.djangoproject.com/en/1.9/ref/validators/
    """
    validator = URLValidator()

    for line in value.splitlines():
        validator(line)


def add_colons(value, pad='0'):
    """Add colons after every second digit.

    This function is used in functions to prettify serials.

    >>> add_colons('teststring')
    'te:st:st:ri:ng'

    Parameters
    ----------

    s : str
        The string to add colons to
    pad : str, default
        If not None, pad the string so that the last element always has two characters. The default is
        ``"0"``.
    """

    if len(value) % 2 == 1 and pad is not None:
        value = '%s%s' % (pad, value)

    return ':'.join([value[i:i + 2] for i in range(0, len(value), 2)])


def int_to_hex(i):
    """Create a hex-representation of the given serial.

    >>> int_to_hex(12345678)
    'BC614E'
    """
    return hex(i)[2:].upper()


def bytes_to_hex(value):
    """Convert a bytes array to hex.

    >>> bytes_to_hex(b'test')
    '74:65:73:74'
    """
    return add_colons(binascii.hexlify(value).upper().decode('utf-8'))


def hex_to_bytes(value):
    """Convert a hex number to bytes.

    This should be the inverse of :py:func:`~django_ca.utils.bytes_to_hex`.

    >>> hex_to_bytes('74:65:73:74')
    b'test'
    """
    return binascii.unhexlify(value.replace(':', ''))


def sanitize_serial(value):
    """Sanitize a serial provided by user/untrusted input.

    This function is intended to be used to get a serial as used internaly by **django-ca** from untrusted
    user input. Internally, serials are stored in upper case and without ``:`` and leading zeros, but user
    output adds at least ``:``.

    Examples:

        >>> sanitize_serial('01:aB')
        '1AB'
    """

    serial = value.upper().replace(':', '')
    if serial != '0':
        serial = serial.lstrip('0')
    if re.search('[^0-9A-F]', serial):
        raise ValueError('%s: Serial has invalid characters' % value)
    return serial


def parse_name(name):
    """Parses a subject string as used in OpenSSLs command line utilities.

    The ``name`` is expected to be close to the subject format commonly used by OpenSSL, for example
    ``/C=AT/L=Vienna/CN=example.com/emailAddress=user@example.com``. The function does its best to be lenient
    on deviations from the format, object identifiers are case-insensitive (e.g. ``cn`` is the same as ``CN``,
    whitespace at the start and end is stripped and the subject does not have to start with a slash (``/``).

    >>> parse_name('/CN=example.com')
    [('CN', 'example.com')]
    >>> parse_name('c=AT/l= Vienna/o="ex org"/CN=example.com')
    [('C', 'AT'), ('L', 'Vienna'), ('O', 'ex org'), ('CN', 'example.com')]

    Dictionary keys are normalized to the values of :py:const:`OID_NAME_MAPPINGS` and keys will be sorted
    based on x509 name specifications regardless of the given order:

    >>> parse_name('L="Vienna / District"/EMAILaddress=user@example.com')
    [('L', 'Vienna / District'), ('emailAddress', 'user@example.com')]
    >>> parse_name('/C=AT/CN=example.com') == parse_name('/CN=example.com/C=AT')
    True

    Due to the magic of :py:const:`NAME_RE`, the function even supports quoting strings and including slashes,
    so strings like ``/OU="Org / Org Unit"/CN=example.com`` will work as expected.

    >>> parse_name('L="Vienna / District"/CN=example.com')
    [('L', 'Vienna / District'), ('CN', 'example.com')]

    But note that it's still easy to trick this function, if you really want to. The following example is
    *not* a valid subject, the location is just bogus, and whatever you were expecting as output, it's
    certainly different:

    >>> parse_name('L="Vienna " District"/CN=example.com')
    [('L', 'Vienna'), ('CN', 'example.com')]

    Examples of where this string is used are:

    .. code-block:: console

        # openssl req -new -key priv.key -out csr -utf8 -batch -sha256 -subj '/C=AT/CN=example.com'
        # openssl x509 -in cert.pem -noout -subject -nameopt compat
        /C=AT/L=Vienna/CN=example.com
    """
    name = name.strip()
    if not name:  # empty subjects are ok
        return []

    try:
        items = [(NAME_CASE_MAPPINGS[t[0].upper()], force_str(t[2])) for t in NAME_RE.findall(name)]
    except KeyError as e:
        raise ValueError('Unknown x509 name field: %s' % e.args[0]) from e

    # Check that no OIDs not in MULTIPLE_OIDS occur more then once
    for key, oid in NAME_OID_MAPPINGS.items():
        if sum(1 for t in items if t[0] == key) > 1 and oid not in MULTIPLE_OIDS:
            raise ValueError('Subject contains multiple "%s" fields' % key)

    return sort_name(items)


def x509_name(name):
    """Parses a subject into a :py:class:`x509.Name <cg:cryptography.x509.Name>`.

    If ``name`` is a string, :py:func:`parse_name` is used to parse it.

    >>> x509_name('/C=AT/CN=example.com')
    <Name(C=AT,CN=example.com)>
    >>> x509_name([('C', 'AT'), ('CN', 'example.com')])
    <Name(C=AT,CN=example.com)>
    """
    if isinstance(name, str):
        name = parse_name(name)

    return x509.Name([x509.NameAttribute(NAME_OID_MAPPINGS[typ], force_str(value)) for typ, value in name])


def x509_relative_name(name):
    """Parse a relative name (RDN) into a :py:class:`~cg:cryptography.x509.RelativeDistinguishedName`.

    >>> x509_relative_name('/CN=example.com')
    <RelativeDistinguishedName(CN=example.com)>
    >>> x509_relative_name([('CN', 'example.com')])
    <RelativeDistinguishedName(CN=example.com)>
    """
    if isinstance(name, x509.RelativeDistinguishedName):
        return name
    if isinstance(name, str):
        name = parse_name(name)

    return x509.RelativeDistinguishedName([
        x509.NameAttribute(NAME_OID_MAPPINGS[typ], force_str(value)) for typ, value in name
    ])


def validate_email(addr):
    """Validate an email address.

    This function raises ``ValueError`` if the email address is not valid.

    >>> validate_email('foo@bar.com')
    'foo@bar.com'
    >>> validate_email('foo@bar com')
    Traceback (most recent call last):
        ...
    ValueError: Invalid domain: bar com

    """
    if '@' not in addr:
        raise ValueError('Invalid email address: %s' % addr)

    node, domain = addr.rsplit('@', 1)
    try:
        domain = idna.encode(domain).decode('utf-8')
    except idna.core.IDNAError as e:
        raise ValueError('Invalid domain: %s' % domain) from e

    return '%s@%s' % (node, domain)


def validate_hostname(hostname, allow_port=False):
    """Validate a hostname, optionally with a given port.

    >>> validate_hostname('example.com')
    'example.com'
    >>> validate_hostname('example.com:8000', allow_port=True)
    'example.com:8000'

    Parameters
    ----------

    hostname : str
        The hostname to validate.
    allow_port : bool, optional
        If ``True``, the hostname can also contain an optional port number, e.g. "example.com:8000".

    Raises
    ------

    ValueError
        If hostname or port are not valid.

    """
    port = None
    if allow_port is True and ':' in hostname:
        hostname, port = hostname.rsplit(':', 1)

        try:
            port = int(port)
        except ValueError as e:
            raise ValueError('%s: Port must be an integer' % port) from e
        else:
            if port < 1 or port > 65535:
                raise ValueError('%s: Port must be between 1 and 65535' % port)

    try:
        encoded = idna.encode(hostname).decode('utf-8')
    except idna.IDNAError as e:
        raise ValueError('%s: Not a valid hostname' % hostname) from e

    if allow_port is True and port is not None:
        return '%s:%s' % (encoded, port)
    return encoded


def validate_key_parameters(key_size=None, key_type='RSA', ecc_curve=None):
    """Validate parameters for private key generation and return sanitized values.

    This function can be used to fail early if invalid parameters are passed, before the private key is
    generated.

    >>> validate_key_parameters()  # defaults
    (1024, 'RSA', None)
    >>> validate_key_parameters(4096, 'ECC', None)  # doctest: +ELLIPSIS
    (None, 'ECC', <cryptography.hazmat.primitives.asymmetric.ec.SECP256R1 object at ...>)
    >>> validate_key_parameters(4000, 'RSA', None)
    Traceback (most recent call last):
        ...
    ValueError: 4000: Key size must be a power of two
    """
    if key_type is None:
        key_type = 'RSA'

    if key_type == 'ECC':
        key_size = None
        ecc_curve = parse_key_curve(ecc_curve)
    elif key_type in ['DSA', 'RSA']:
        if key_size is None:
            key_size = ca_settings.CA_DEFAULT_KEY_SIZE

        if not is_power2(key_size):
            raise ValueError("%s: Key size must be a power of two" % key_size)
        if key_size < ca_settings.CA_MIN_KEY_SIZE:
            raise ValueError("%s: Key size must be least %s bits" % (
                key_size, ca_settings.CA_MIN_KEY_SIZE))
    else:
        raise ValueError('%s: Unknown key type' % key_type)

    return key_size, key_type, ecc_curve


def generate_private_key(key_size, key_type, ecc_curve):
    """Generate a private key.

    This function assumes that you called :py:func:`~django_ca.utils.validate_key_parameters` on the input
    values and does not do any sanity checks on its own.

    Parameters
    ----------

    key_size : int
        The size of the private key (not used for ECC keys).
    key_type : {'RSA', 'DSA', 'ECC'}
        The type of the private key.
    ecc_curve : :py:class:`~cg:cryptography.hazmat.primitives.asymmetric.ec.EllipticCurve`
        The ECC curve to use for an ECC key.

    Returns
    -------

    key
        A private key of the appropriate type.
    """
    if key_type == 'DSA':
        private_key = dsa.generate_private_key(key_size=key_size, backend=default_backend())
    elif key_type == 'ECC':
        private_key = ec.generate_private_key(ecc_curve, default_backend())
    else:
        private_key = rsa.generate_private_key(public_exponent=65537, key_size=key_size,
                                               backend=default_backend())

    return private_key


def parse_general_name(name):
    """Parse a general name from user input.

    This function will do its best to detect the intended type of any value passed to it:

    >>> parse_general_name('example.com')
    <DNSName(value='example.com')>
    >>> parse_general_name('*.example.com')
    <DNSName(value='*.example.com')>
    >>> parse_general_name('.example.com')  # Syntax used e.g. for NameConstraints: All levels of subdomains
    <DNSName(value='.example.com')>
    >>> parse_general_name('user@example.com')
    <RFC822Name(value='user@example.com')>
    >>> parse_general_name('https://example.com')
    <UniformResourceIdentifier(value='https://example.com')>
    >>> parse_general_name('1.2.3.4')
    <IPAddress(value=1.2.3.4)>
    >>> parse_general_name('fd00::1')
    <IPAddress(value=fd00::1)>
    >>> parse_general_name('/CN=example.com')
    <DirectoryName(value=<Name(CN=example.com)>)>

    The default fallback is to assume a :py:class:`~cg:cryptography.x509.DNSName`. If this doesn't
    work, an exception will be raised:

    >>> parse_general_name('foo..bar`*123')  # doctest: +ELLIPSIS
    Traceback (most recent call last):
        ...
    ValueError: Could not parse name: foo..bar`*123

    If you want to override detection, you can prefix the name to match :py:const:`GENERAL_NAME_RE`:

    >>> parse_general_name('email:user@example.com')
    <RFC822Name(value='user@example.com')>
    >>> parse_general_name('URI:https://example.com')
    <UniformResourceIdentifier(value='https://example.com')>
    >>> parse_general_name('dirname:/CN=example.com')
    <DirectoryName(value=<Name(CN=example.com)>)>

    Some more exotic values can only be generated by using this prefix:

    >>> parse_general_name('rid:2.5.4.3')
    <RegisteredID(value=<ObjectIdentifier(oid=2.5.4.3, name=commonName)>)>
    >>> parse_general_name('otherName:2.5.4.3;UTF8:example.com')
    <OtherName(type_id=<ObjectIdentifier(oid=2.5.4.3, name=commonName)>, value=b'example.com')>

    If you give a prefixed value, this function is less forgiving of any typos and does not catch any
    exceptions:

    >>> parse_general_name('email:foo@bar com')
    Traceback (most recent call last):
        ...
    ValueError: Invalid domain: bar com

    """
    # pylint: disable=too-many-return-statements,too-many-branches,too-many-statements

    if isinstance(name, x509.GeneralName):
        return name
    if not isinstance(name, str):
        raise ValueError('Cannot parse general name %s: Must be of type str (was: %s).' %
                         (name, type(name).__name__))

    typ = None
    match = GENERAL_NAME_RE.match(name)
    if match is not None:
        typ, name = match.groups()
        typ = typ.lower()

    if typ is None:
        if re.match('[a-z0-9]{2,}://', name):  # Looks like a URI
            try:
                return x509.UniformResourceIdentifier(encode_url(name))
            except idna.IDNAError:
                pass

        if '@' in name:  # Looks like an Email address
            try:
                return x509.RFC822Name(validate_email(name))
            except ValueError:
                pass

        if name.strip().startswith('/'):  # maybe it's a dirname?
            return x509.DirectoryName(x509_name(name))

        # Try to parse this as IPAddress/Network
        try:
            return x509.IPAddress(ip_address(name))
        except ValueError:
            pass
        try:
            return x509.IPAddress(ip_network(name))
        except ValueError:
            pass

        # Almost anything passes as DNS name, so this is our default fallback
        try:
            return x509.DNSName(encode_dns(name))
        except idna.IDNAError as e:
            raise ValueError('Could not parse name: %s' % name) from e

    if typ == 'uri':
        try:
            return x509.UniformResourceIdentifier(encode_url(name))
        except idna.IDNAError as e:
            raise ValueError('Could not parse DNS name in URL: %s' % name) from e
    elif typ == 'email':
        return x509.RFC822Name(validate_email(name))  # validate_email already raises ValueError
    elif typ == 'ip':
        try:
            return x509.IPAddress(ip_address(name))
        except ValueError:
            pass

        try:
            return x509.IPAddress(ip_network(name))
        except ValueError:
            pass

        raise ValueError('Could not parse IP address.')
    elif typ == 'rid':
        return x509.RegisteredID(x509.ObjectIdentifier(name))
    elif typ == 'othername':
        regex = "(.*);(.*):(.*)"
        if re.match(regex, name) is not None:
            oid, asn_typ, val = re.match(regex, name).groups()
            oid = x509.ObjectIdentifier(oid)
            if asn_typ == 'UTF8':
                val = val.encode('utf-8')
            elif asn_typ == 'OctetString':
                val = bytes(bytearray.fromhex(val))
                val = OctetString(val).dump()
            else:
                raise ValueError('Unsupported ASN type in otherName: %s' % asn_typ)
            val = force_bytes(val)
            return x509.OtherName(oid, val)

        raise ValueError('Incorrect otherName format: %s' % name)
    elif typ == 'dirname':
        return x509.DirectoryName(x509_name(name))
    else:
        try:
            return x509.DNSName(encode_dns(name))
        except idna.IDNAError as e:
            raise ValueError('Could not parse DNS name: %s' % name) from e


def parse_hash_algorithm(value=None):
    """Parse a hash algorithm value.

    The most common use case is to pass a str naming a class in
    :py:mod:`~cg:cryptography.hazmat.primitives.hashes`.

    For convenience, passing ``None`` will return the value of :ref:`CA_DIGEST_ALGORITHM
    <settings-ca-digest-algorithm>`, and passing an
    :py:class:`~cg:cryptography.hazmat.primitives.hashes.HashAlgorithm` will return that
    instance unchanged.

    Example usage::

        >>> parse_hash_algorithm()  # doctest: +ELLIPSIS
        <cryptography.hazmat.primitives.hashes.SHA512 object at ...>
        >>> parse_hash_algorithm('SHA512')  # doctest: +ELLIPSIS
        <cryptography.hazmat.primitives.hashes.SHA512 object at ...>
        >>> parse_hash_algorithm(' SHA512 ')  # doctest: +ELLIPSIS
        <cryptography.hazmat.primitives.hashes.SHA512 object at ...>
        >>> parse_hash_algorithm(hashes.SHA512)  # doctest: +ELLIPSIS
        <cryptography.hazmat.primitives.hashes.SHA512 object at ...>
        >>> parse_hash_algorithm(hashes.SHA512())  # doctest: +ELLIPSIS
        <cryptography.hazmat.primitives.hashes.SHA512 object at ...>
        >>> parse_hash_algorithm('Wrong')  # doctest: +ELLIPSIS
        Traceback (most recent call last):
            ...
        ValueError: Unknown hash algorithm: Wrong
        >>> parse_hash_algorithm(object())  # doctest: +ELLIPSIS
        Traceback (most recent call last):
            ...
        ValueError: Unknown type passed: object

    Parameters
    ----------

    value : str or :py:class:`~cg:cryptography.hazmat.primitives.hashes.HashAlgorithm`, optional
        The value to parse, the function description on how possible values are used.

    Returns
    -------

    algorithm
        A :py:class:`~cg:cryptography.hazmat.primitives.hashes.HashAlgorithm` instance.

    Raises
    ------

    ValueError
        If an unknown object is passed or if ``value`` does not name a known algorithm.
    """
    if value is None:
        return ca_settings.CA_DIGEST_ALGORITHM
    if isinstance(value, type) and issubclass(value, hashes.HashAlgorithm):
        return value()
    if isinstance(value, hashes.HashAlgorithm):
        return value
    if isinstance(value, str):
        try:
            return getattr(hashes, value.strip())()
        except AttributeError as e:
            raise ValueError('Unknown hash algorithm: %s' % value) from e

    raise ValueError('Unknown type passed: %s' % type(value).__name__)


def parse_encoding(value=None):
    """Parse a value to a valid encoding.

    This function accepts either a member of
    :py:class:`~cg:cryptography.hazmat.primitives.serialization.Encoding` or a string describing a member. If
    no value is passed, it will assume ``PEM`` as a default value. Note that ``"ASN1"`` is treated as an alias
    for ``"DER"``.

        >>> parse_encoding()
        <Encoding.PEM: 'PEM'>
        >>> parse_encoding('DER')
        <Encoding.DER: 'DER'>
        >>> parse_encoding(Encoding.PEM)
        <Encoding.PEM: 'PEM'>
    """
    if value is None:
        return ca_settings.CA_DEFAULT_ENCODING
    if isinstance(value, Encoding):
        return value
    if isinstance(value, str):
        if value == 'ASN1':
            value = 'DER'

        try:
            return getattr(Encoding, value)
        except AttributeError as e:
            raise ValueError('Unknown encoding: %s' % value) from e

    raise ValueError('Unknown type passed: %s' % type(value).__name__)


def parse_key_curve(value=None):
    """Parse an elliptic curve value.

    This function uses a value identifying an elliptic curve to return an
    :py:class:`~cg:cryptography.hazmat.primitives.asymmetric.ec.EllipticCurve` instance. The name must match a
    class name of one of the classes named under "Elliptic Curves" in
    :any:`cg:hazmat/primitives/asymmetric/ec`.

    For convenience, passing ``None`` will return the value of :ref:`CA_DEFAULT_ECC_CURVE
    <settings-ca-default-ecc-curve>`, and passing an
    :py:class:`~cg:cryptography.hazmat.primitives.asymmetric.ec.EllipticCurve` will return that instance
    unchanged.

    Example usage::

        >>> parse_key_curve('SECP256R1')  # doctest: +ELLIPSIS
        <cryptography.hazmat.primitives.asymmetric.ec.SECP256R1 object at ...>
        >>> parse_key_curve('SECP384R1')  # doctest: +ELLIPSIS
        <cryptography.hazmat.primitives.asymmetric.ec.SECP384R1 object at ...>
        >>> parse_key_curve(ec.SECP256R1())  # doctest: +ELLIPSIS
        <cryptography.hazmat.primitives.asymmetric.ec.SECP256R1 object at ...>
        >>> parse_key_curve()  # doctest: +ELLIPSIS
        <cryptography.hazmat.primitives.asymmetric.ec.SECP256R1 object at ...>

    Parameters
    ----------

    value : str, otional
        The name of the curve or ``None`` to return the default curve.

    Returns
    -------

    curve
        An :py:class:`~cg:cryptography.hazmat.primitives.asymmetric.ec.EllipticCurve` instance.

    Raises
    ------

    ValueError
        If the named curve is not supported.
    """
    if isinstance(value, ec.EllipticCurve):
        return value  # name was already parsed
    if value is None:
        return ca_settings.CA_DEFAULT_ECC_CURVE

    curve = getattr(ec, value.strip(), type)
    if not issubclass(curve, ec.EllipticCurve):
        raise ValueError('%s: Not a known Eliptic Curve' % value)
    return curve()


def get_cert_builder(expires, serial=None):
    """Get a basic X509 cert builder object.

    .. TODO:: deprecate support for passing datetime as expires

    Parameters
    ----------

    expires : datetime or timedelta
        When this certificate will expire.
    serial : int, optional
        Serial number to set for this certificate. Use :py:func:`~cg:cryptography.x509.random_serial_number`
        to generate such a value. By default, a value will be generated.
    """
    now = datetime.utcnow().replace(second=0, microsecond=0)

    if serial is None:
        serial = x509.random_serial_number()
    if expires is None:
        expires = now + ca_settings.CA_DEFAULT_EXPIRES
    elif isinstance(expires, timedelta):
        expires = now + expires
    else:
        expires = expires.replace(second=0, microsecond=0)

    builder = x509.CertificateBuilder()
    builder = builder.not_valid_before(now)
    builder = builder.not_valid_after(expires)
    builder = builder.serial_number(serial)

    return builder


def read_file(path):
    """Read the file from the given path.

    If ``path`` is an absolute path, reads a file from the local filesystem. For relative paths, read the file
    using the storage backend configured using :ref:`CA_FILE_STORAGE <settings-ca-file-storage>`.
    """
    if os.path.isabs(path):
        with open(path, 'rb') as stream:
            return stream.read()

    stream = ca_storage.open(path)

    try:
        # NOTE: In the python:3.9-rc-alpine3.10 Docker image, this is marked as a missed branch :-(
        return stream.read()  # pragma: no branch
    finally:
        stream.close()


# Note used currently, but left here for future reference
#def write_private_file(path, data):
#    """Function to write binary data to a file that will only be readable to the user."""
#
#    with os.fdopen(os.open(path, os.O_CREAT | os.O_WRONLY, 0o400), 'wb') as fh:
#        fh.write(data)


def shlex_split(val, sep):
    """Split a character on the given set of characters.

    Example::

        >>> shlex_split('foo,bar', ', ')
        ['foo', 'bar']
        >>> shlex_split('foo\\\\,bar1', ',')  # escape a separator
        ['foo,bar1']
        >>> shlex_split('"foo,bar", bla', ', ')
        ['foo,bar', 'bla']
        >>> shlex_split('foo,"bar,bla"', ',')
        ['foo', 'bar,bla']
    """
    lex = shlex.shlex(val, posix=True)
    lex.whitespace = sep
    lex.whitespace_split = True
    return list(lex)


class GeneralNameList(list):
    """List that holds :py:class:`~cg:cryptography.x509.GeneralName` instances and parses ``str`` when added.

    A ``GeneralNameList`` is a ``list`` subclass that will always only hold
    :py:class:`~cg:cryptography.x509.GeneralName` instances, but any ``str`` passed to it will be passed to
    :py:func:`~django_ca.utils.parse_general_name`::

        >>> from cryptography import x509
        >>> l = GeneralNameList(['example.com'])
        >>> l += ['DNS:example.net', x509.DNSName('example.org')]
        >>> print(l)
        <GeneralNameList: ['DNS:example.com', 'DNS:example.net', 'DNS:example.org']>
        >>> 'example.com' in l, 'DNS:example.com' in l, x509.DNSName('example.com') in l
        (True, True, True)
        >>> l == ['example.com', 'example.net', 'example.org']
        True
        >>> l == [x509.DNSName('example.com'), 'example.net', 'DNS:example.org']
        True

    """
    def __init__(self, iterable=tuple()):
        if isinstance(iterable, (str, x509.GeneralName)):
            iterable = [iterable]

        super().__init__(parse_general_name(v) for v in iterable)

    def serialize(self):
        """Generate a list of formatted names."""
        for val in self:
            yield format_general_name(val)

    def __add__(self, value):  # self + other_list
        if isinstance(value, GeneralNameList) is False:
            value = GeneralNameList(value)
        return GeneralNameList(list(self) + list(value))

    def __contains__(self, value):  # value in self
        try:
            value = parse_general_name(value)
        except ValueError:
            return False

        return list.__contains__(self, value)

    def __eq__(self, other):  # value == other
        if isinstance(other, GeneralNameList) is False and isinstance(other, list) is True:
            other = GeneralNameList(other)
        return list.__eq__(self, other)

    def __iadd__(self, value):  # self += value
        return list.__iadd__(self, (parse_general_name(v) for v in value))

    def __repr__(self):
        return '<GeneralNameList: %r>' % [format_general_name(v) for v in self]

    def __setitem__(self, key, value):  # l[0] = 'example.com'
        if isinstance(key, slice):  # l[0:1] = ['example.com']
            list.__setitem__(self, key, (parse_general_name(v) for v in value))
        else:
            list.__setitem__(self, key, parse_general_name(value))

    def append(self, o):
        list.append(self, parse_general_name(o))

    def count(self, value):
        try:
            value = parse_general_name(value)
        except ValueError:
            return 0

        return list.count(self, value)

    def extend(self, iterable):
        list.extend(self, (parse_general_name(i) for i in iterable))

    def index(self, value, *args):
        return list.index(self, parse_general_name(value), *args)

    def insert(self, index, o):
        list.insert(self, index, parse_general_name(o))

    def remove(self, value):
        list.remove(self, parse_general_name(value))


def get_crl_cache_key(serial, algorithm=hashes.SHA512, encoding=Encoding.DER, scope=None):
    """Function to get a cache key for a CRL with the given parameters."""

    return 'crl_%s_%s_%s_%s' % (serial, algorithm.name, encoding.name, scope)


ca_storage = get_storage_class(ca_settings.CA_FILE_STORAGE)(**ca_settings.CA_FILE_STORAGE_KWARGS)



# Generated by Django 2.1.3 on 2018-11-28 20:50

import datetime
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_ca', '0008_auto_20171203_2001'),
    ]

    operations = [
        migrations.AddField(
            model_name='certificate',
            name='valid_from',
            field=models.DateTimeField(default=datetime.datetime(2018, 11, 28, 20, 50, 23, 447354)),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='certificateauthority',
            name='valid_from',
            field=models.DateTimeField(default=datetime.datetime(2018, 11, 28, 20, 50, 47, 196936)),
            preserve_default=False,
        ),
    ]



# This file is part of django-ca (https://github.com/mathiasertl/django-ca).
#
# django-ca is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# django-ca is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with django-ca.  If not,
# see <http://www.gnu.org/licenses/>.

"""Management command to write a CAs public key to stdout or a file.

.. seealso:: https://docs.djangoproject.com/en/dev/howto/custom-management-commands/
"""

from cryptography.hazmat.primitives.serialization import Encoding

from django.core.management.base import CommandError

from ..base import BaseCommand


class Command(BaseCommand):  # pylint: disable=missing-class-docstring
    help = "Dump a certificate authority to a file."
    binary_output = True

    def add_arguments(self, parser):
        super().add_arguments(parser)
        self.add_format(parser)
        self.add_ca(parser, arg='ca', allow_disabled=True)
        parser.add_argument('-b', '--bundle', default=False, action='store_true',
                            help='Dump the whole certificate bundle.')
        parser.add_argument('path', nargs='?', default='-',
                            help='Path where to dump the certificate. Use "-" for stdout.')

    def handle(self, ca, path, **options):  # pylint: disable=arguments-differ
        if options['bundle'] and options['format'] == Encoding.DER:
            raise CommandError('Cannot dump bundle when using DER format.')

        if options['bundle']:
            certs = ca.bundle
        else:
            certs = [ca]

        self.dump(path, b''.join([c.dump_certificate(options['format']) for c in certs]))



# This file is part of django-ca (https://github.com/mathiasertl/django-ca).
#
# django-ca is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# django-ca is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with django-ca.  If not,
# see <http://www.gnu.org/licenses/>.

import argparse
import getpass
import sys
from datetime import timedelta
from textwrap import indent

from cryptography import x509
from cryptography.hazmat.primitives.serialization import Encoding

from django.core.exceptions import ImproperlyConfigured
from django.core.management.base import BaseCommand as _BaseCommand
from django.core.management.base import CommandError
from django.core.management.base import OutputWrapper
from django.core.management.color import no_style
from django.core.validators import URLValidator
from django.utils import timezone
from django.utils.encoding import force_bytes

from .. import ca_settings
from ..constants import ReasonFlags
from ..extensions import ExtendedKeyUsage
from ..extensions import Extension
from ..extensions import IssuerAlternativeName
from ..extensions import KeyUsage
from ..extensions import NullExtension
from ..extensions import SubjectAlternativeName
from ..extensions import TLSFeature
from ..models import Certificate
from ..models import CertificateAuthority
from ..subject import Subject
from ..utils import SUBJECT_FIELDS
from ..utils import add_colons
from ..utils import is_power2
from ..utils import parse_encoding
from ..utils import parse_hash_algorithm
from ..utils import parse_key_curve
from ..utils import shlex_split


class SubjectAction(argparse.Action):
    def __call__(self, parser, namespace, value, option_string=None):
        try:
            value = Subject(value)
        except ValueError as e:
            parser.error(e)
        setattr(namespace, self.dest, value)


class FormatAction(argparse.Action):
    def __call__(self, parser, namespace, value, option_string=None):
        try:
            value = parse_encoding(value)
        except ValueError as e:
            parser.error(str(e))

        setattr(namespace, self.dest, value)


class AlgorithmAction(argparse.Action):
    def __call__(self, parser, namespace, value, option_string=None):
        try:
            value = parse_hash_algorithm(value)
        except ValueError as e:
            parser.error(str(e))

        setattr(namespace, self.dest, value)


class KeyCurveAction(argparse.Action):
    """Action to parse an ECC curve value."""

    def __call__(self, parser, namespace, value, option_string=None):

        try:
            curve = parse_key_curve(value)
        except ValueError as e:
            parser.error(e)
        setattr(namespace, self.dest, curve)


class KeySizeAction(argparse.Action):
    def __call__(self, parser, namespace, value, option_string=None):
        option_string = option_string or 'key size'

        if not is_power2(value):
            parser.error('%s must be a power of two (2048, 4096, ...)' % option_string)
        elif value < ca_settings.CA_MIN_KEY_SIZE:
            parser.error('%s must be at least %s bits.'
                         % (option_string, ca_settings.CA_MIN_KEY_SIZE))
        setattr(namespace, self.dest, value)


class PasswordAction(argparse.Action):
    def __init__(self, prompt=None, **kwargs):
        super(PasswordAction, self).__init__(**kwargs)
        self.prompt = prompt

    def __call__(self, parser, namespace, value, option_string=None):
        if value is None:
            kwargs = {}
            if self.prompt:
                kwargs['prompt'] = self.prompt
            value = getpass.getpass(**kwargs)

        setattr(namespace, self.dest, value.encode('utf-8'))


class CertificateAction(argparse.Action):
    def __init__(self, allow_revoked=False, **kwargs):
        super(CertificateAction, self).__init__(**kwargs)
        self.allow_revoked = allow_revoked

    def __call__(self, parser, namespace, value, option_string=None):
        queryset = Certificate.objects.all()
        if self.allow_revoked is False:
            queryset = queryset.filter(revoked=False)

        try:
            setattr(namespace, self.dest, queryset.get_by_serial_or_cn(value))
        except Certificate.DoesNotExist:
            raise parser.error('%s: Certificate not found.' % value)
        except Certificate.MultipleObjectsReturned:
            raise parser.error('%s: Multiple certificates match.' % value)


class CertificateAuthorityAction(argparse.Action):
    def __init__(self, allow_disabled=False, allow_unusable=False, **kwargs):
        super(CertificateAuthorityAction, self).__init__(**kwargs)
        self.allow_disabled = allow_disabled
        self.allow_unusable = allow_unusable

    def __call__(self, parser, namespace, value, option_string=None):
        qs = CertificateAuthority.objects.all()
        if self.allow_disabled is False:
            qs = qs.enabled()

        try:
            value = qs.get_by_serial_or_cn(value)
        except CertificateAuthority.DoesNotExist:
            parser.error('%s: Certificate authority not found.' % value)
        except CertificateAuthority.MultipleObjectsReturned:
            parser.error('%s: Multiple Certificate authorities match.' % value)

        # verify that the private key exists
        if not self.allow_unusable and not value.key_exists:
            parser.error('%s: %s: Private key does not exist.' % (value, value.private_key_path))

        setattr(namespace, self.dest, value)


class URLAction(argparse.Action):
    def __call__(self, parser, namespace, value, option_string=None):
        validator = URLValidator()
        try:
            validator(value)
        except Exception:
            parser.error('%s: Not a valid URL.' % value)
        setattr(namespace, self.dest, value)


def parse_timedelta(value):
    try:
        value = int(value)
    except Exception:
        raise argparse.ArgumentTypeError('Value must be an integer: "%s"' % value)
    if value <= 0:
        raise argparse.ArgumentTypeError('Value must not be negative.')

    return timedelta(days=value)


class ExpiresAction(argparse.Action):
    def __init__(self, *args, **kwargs):
        kwargs['type'] = parse_timedelta
        kwargs.setdefault('default', ca_settings.CA_DEFAULT_EXPIRES)
        super(ExpiresAction, self).__init__(*args, **kwargs)

    def __call__(self, parser, namespace, value, option_string=None):
        setattr(namespace, self.dest, value)


class MultipleURLAction(argparse.Action):
    def __call__(self, parser, namespace, value, option_string=None):
        validator = URLValidator()
        try:
            validator(value)
        except Exception:
            parser.error('%s: Not a valid URL.' % value)

        if getattr(namespace, self.dest) is None:
            setattr(namespace, self.dest, [])

        getattr(namespace, self.dest).append(value)


class ExtensionAction(argparse.Action):
    def __init__(self, *args, **kwargs):
        self.extension = kwargs.pop('extension')
        kwargs['dest'] = self.extension.key
        super(ExtensionAction, self).__init__(*args, **kwargs)


class OrderedSetExtensionAction(ExtensionAction):
    def __call__(self, parser, namespace, value, option_string=None):
        ext = self.extension()

        values = shlex_split(value, ', ')
        if values[0] == 'critical':
            values = values[1:]
            ext.critical = True
        else:
            ext.critical = False

        try:
            ext |= values
        except ValueError as e:
            parser.error('Invalid extension value: %s: %s' % (value, e))

        setattr(namespace, self.dest, ext)


class AlternativeNameAction(ExtensionAction):
    def __call__(self, parser, namespace, value, option_string=None):
        setattr(namespace, self.dest, self.extension({'value': [value]}))


class ReasonAction(argparse.Action):
    def __init__(self, *args, **kwargs):
        kwargs['choices'] = sorted([r.name for r in ReasonFlags])
        super(ReasonAction, self).__init__(*args, **kwargs)

    def __call__(self, parser, namespace, value, option_string=None):
        # NOTE: set of choices already assures that value is a valid ReasonFlag
        setattr(namespace, self.dest, ReasonFlags[value])


class BinaryOutputWrapper(OutputWrapper):
    def __init__(self, out, ending=b'\n'):
        super(BinaryOutputWrapper, self).__init__(out, ending=ending)

    def write(self, msg, style_func=None, ending=None):
        ending = self.ending if ending is None else ending
        msg = force_bytes(msg)

        if ending and not msg.endswith(ending):
            msg += ending
        self._out.write(msg)


class BaseCommand(_BaseCommand):
    binary_output = False

    def __init__(self, stdout=None, stderr=None, no_color=False):
        if self.binary_output is True:
            self.stdout = BinaryOutputWrapper(stdout or sys.stdout.buffer)
            self.stderr = BinaryOutputWrapper(stderr or sys.stderr.buffer)
            self.style = no_style()
        else:
            super(BaseCommand, self).__init__(stdout, stderr, no_color=no_color)

    def dump(self, path, data):
        """Dump `data` to `path` (``-`` means stdout)."""

        if path == '-':
            self.stdout.write(data, ending=b'')
        else:
            try:
                with open(path, 'wb') as stream:
                    stream.write(data)
            except IOError as ex:
                raise CommandError(ex) from ex

    def execute(self, *args, **options):
        if self.binary_output is True:
            if options.get('stdout'):  # pragma: no branch
                self.stdout = BinaryOutputWrapper(options.pop('stdout'))
            if options.get('stderr'):  # pragma: no branch
                self.stderr = BinaryOutputWrapper(options.pop('stderr'))
            options['no_color'] = True

        super(BaseCommand, self).execute(*args, **options)

    def add_algorithm(self, parser):
        """Add the --algorithm option."""

        help = 'The HashAlgorithm that will be used to generate the signature (default: %(default)s).' % {
            'default': ca_settings.CA_DIGEST_ALGORITHM.name, }

        parser.add_argument(
            '--algorithm', metavar='{sha512,sha256,...}', default=ca_settings.CA_DIGEST_ALGORITHM,
            action=AlgorithmAction, help=help)

    @property
    def valid_subject_keys(self):
        fields = ['"%s"' % f for f in SUBJECT_FIELDS]
        return '%s and %s' % (', '.join(fields[:-1]), fields[-1])

    def add_subject(self, parser, arg='subject', metavar=None, help=None):
        parser.add_argument(arg, action=SubjectAction, metavar=metavar, help=help)

    def add_ca(self, parser, arg='--ca', help='Certificate authority to use (default: %(default)s).',
               allow_disabled=False, no_default=False, allow_unusable=False):
        """Add the ``--ca`` action.

        Parameters
        ----------

        parser
        arg : str, optional
        help : str, optional
        allow_disabled : bool, optional
        no_default : bool, optional
        allow_unusable : bool, optional
        """
        if no_default is True:
            default = None
        else:
            try:
                default = CertificateAuthority.objects.default()
            except ImproperlyConfigured:
                default = None

        help = help % {'default': add_colons(default.serial) if default else None}
        parser.add_argument('%s' % arg, metavar='SERIAL', help=help, default=default,
                            allow_disabled=allow_disabled, allow_unusable=allow_unusable,
                            action=CertificateAuthorityAction)

    def add_ecc_curve(self, parser):
        curve_help = 'Elliptic Curve used for ECC keys (default: %(default)s).' % {
            'default': ca_settings.CA_DEFAULT_ECC_CURVE.__class__.__name__,
        }
        parser.add_argument('--ecc-curve', metavar='CURVE', action=KeyCurveAction,
                            default=ca_settings.CA_DEFAULT_ECC_CURVE,
                            help=curve_help)

    def add_format(self, parser, default=Encoding.PEM, help_text=None, opts=None):
        """Add the --format option."""

        if opts is None:
            opts = ['-f', '--format']
        if help_text is None:
            help_text = 'The format to use ("ASN1" is an alias for "DER", default: %(default)s).'
        help_text = help_text % {'default': default.name}
        parser.add_argument(*opts, metavar='{PEM,ASN1,DER}', default=default,
                            action=FormatAction, help=help_text)

    def add_key_size(self, parser):
        parser.add_argument(
            '--key-size', type=int, action=KeySizeAction, default=ca_settings.CA_DEFAULT_KEY_SIZE,
            metavar='{2048,4096,8192,...}',
            help="Key size for the private key (default: %(default)s).")

    def add_key_type(self, parser):
        parser.add_argument(
            '--key-type', choices=['RSA', 'DSA', 'ECC'], default='RSA',
            help="Key type for the private key (default: %(default)s).")

    def add_password(self, parser, help=None):
        if help is None:
            help = 'Password used for accessing the private key of the CA.'
        parser.add_argument('-p', '--password', nargs='?', action=PasswordAction, help=help)

    def add_profile(self, parser, help_text):
        group = parser.add_argument_group('profiles', help_text)
        group = group.add_mutually_exclusive_group()
        for name, profile in ca_settings.CA_PROFILES.items():
            group.add_argument('--%s' % name, action='store_const', const=name, dest='profile',
                               help=profile.get('description', ''))

    def indent(self, s, prefix='    '):
        return indent(s, prefix)

    def print_extension(self, ext):
        if isinstance(ext, Extension):
            if isinstance(ext, NullExtension):
                if ext.critical:
                    # NOTE: Only PrecertPoison is ever marked as critical
                    self.stdout.write('%s (critical): Yes' % ext.name)
                else:
                    self.stdout.write('%s: Yes' % ext.name)
            else:
                if ext.critical:
                    self.stdout.write('%s (critical):' % ext.name)
                else:
                    self.stdout.write('%s:' % ext.name)

                self.stdout.write(self.indent(ext.as_text()))
        elif isinstance(ext, x509.Extension):
            if ext.critical:  # pragma: no cover - all unrecognized extensions that we have are non-critical
                self.stdout.write('%s (critical): %s' % (ext.oid._name, ext.oid.dotted_string))
            else:
                self.stdout.write('%s: %s' % (ext.oid._name, ext.oid.dotted_string))
        else:  # pragma: no cover
            raise ValueError('Received unknown extension type: %s' % type(ext))

    def print_extensions(self, cert):
        for ext in cert.extensions:
            self.print_extension(ext)

    def test_private_key(self, ca, password):
        try:
            ca.key(password)
        except Exception as e:
            raise CommandError(str(e))


class BaseSignCommand(BaseCommand):
    sign_extensions = {
        SubjectAlternativeName,
        KeyUsage,
        ExtendedKeyUsage,
        TLSFeature,
    }

    def add_base_args(self, parser, no_default_ca=False):
        self.add_subject_group(parser)
        self.add_algorithm(parser)
        self.add_ca(parser, no_default=no_default_ca)
        self.add_password(parser)
        self.add_extensions(parser)

        parser.add_argument(
            '--expires', default=ca_settings.CA_DEFAULT_EXPIRES, action=ExpiresAction,
            help='Sign the certificate for DAYS days (default: %(default)s)')
        parser.add_argument(
            '--alt', metavar='DOMAIN', action=AlternativeNameAction, extension=SubjectAlternativeName,
            help='Add a subjectAltName to the certificate (may be given multiple times)')
        parser.add_argument(
            '--watch', metavar='EMAIL', action='append', default=[],
            help='Email EMAIL when this certificate expires (may be given multiple times)')
        parser.add_argument(
            '--out', metavar='FILE',
            help='Save signed certificate to FILE. If omitted, print to stdout.')

    def add_subject_group(self, parser):
        group = parser.add_argument_group('Certificate subject', self.subject_help)

        # NOTE: We do not set the default argument here because that would mask the user not
        # setting anything at all.
        self.add_subject(
            group, arg='--subject', metavar='/key1=value1/key2=value2/...',
            help='''Valid keys are %s. Pass an empty value (e.g. "/C=/ST=...") to remove a field
                 from the subject.''' % self.valid_subject_keys)

    def add_extensions(self, parser):
        group = parser.add_argument_group('X509 v3 certificate extensions', self.add_extensions_help)
        group.add_argument(
            '--key-usage', metavar='VALUES', action=OrderedSetExtensionAction, extension=KeyUsage,
            help='The keyUsage extension, e.g. "critical,keyCertSign".')
        group.add_argument(
            '--ext-key-usage', metavar='VALUES', action=OrderedSetExtensionAction, extension=ExtendedKeyUsage,
            help='The extendedKeyUsage extension, e.g. "serverAuth,clientAuth".')
        group.add_argument(
            '--tls-feature', metavar='VALUES', action=OrderedSetExtensionAction, extension=TLSFeature,
            help='TLS Feature extensions.')

    def test_options(self, *args, **options):
        ca = options['ca']
        if ca.expires < timezone.now() + options['expires']:
            max_days = (ca.expires - timezone.now()).days
            raise CommandError(
                'Certificate would outlive CA, maximum expiry for this CA is %s days.' % max_days)

        # See if we can work with the private key
        self.test_private_key(ca, options['password'])


class CertCommand(BaseCommand):
    allow_revoked = False

    def add_arguments(self, parser):
        parser.add_argument(
            'cert', action=CertificateAction, allow_revoked=self.allow_revoked,
            help='''Certificate by CommonName or serial. If you give a CommonName (which is not by
                definition unique) there must be only one valid certificate with the given
                CommonName.''')
        super(CertCommand, self).add_arguments(parser)


class CertificateAuthorityDetailMixin:
    """Mixin to add common arguments to init_ca and edit_ca."""

    def add_general_args(self, parser, default=''):  # pylint: disable=no-self-use
        """Add some general arguments."""

        group = parser.add_argument_group('General', 'General information about the CA.')
        group.add_argument('--caa', default=default, metavar='NAME', help='CAA record for this CA.')
        group.add_argument('--website', default=default, metavar='URL', action=URLAction,
                           help='Browsable URL for the CA.')
        group.add_argument('--tos', default=default, metavar='URL', action=URLAction,
                           help='Terms of service URL for the CA.')

    def add_acme_group(self, parser):  # pylint: disable=no-self-use
        """Add configuration for ACMEv2."""

        if not ca_settings.CA_ENABLE_ACME:
            return

        group = parser.add_argument_group('ACMEv2', 'ACMEv2 configuration.')

        enable_group = group.add_mutually_exclusive_group()
        enable_group.add_argument('--acme-enable', dest='acme_enabled', action='store_true', default=None,
                                  help="Enable ACMEv2 support.")
        enable_group.add_argument('--acme-disable', dest='acme_enabled', action='store_false',
                                  help='Disable ACMEv2 support.')

        disable_group = group.add_mutually_exclusive_group()
        disable_group.add_argument(
            '--acme-contact-optional', dest='acme_requires_contact', action='store_false', default=None,
            help="Do not require email address during ACME account registration.")

        disable_group.add_argument(
            '--acme-contact-required', dest='acme_requires_contact', action='store_true',
            help="Require email address during ACME account registration.")

    def add_ca_args(self, parser):  # pylint: disable=no-self-use
        """Add CA arguments."""

        group = parser.add_argument_group(
            'X509 v3 certificate extensions for signed certificates',
            'Extensions added when signing certificates.')
        group.add_argument('--issuer-url', metavar='URL', action=URLAction,
                           help='URL to the certificate of your CA (in DER format).')
        group.add_argument(
            '--issuer-alt-name', metavar='URL', action=AlternativeNameAction, extension=IssuerAlternativeName,
            help='URL to the homepage of your CA.'
        )
        group.add_argument(
            '--crl-url', metavar='URL', action=MultipleURLAction, default=[],
            help='URL to a certificate revokation list. Can be given multiple times.'
        )
        group.add_argument('--ocsp-url', metavar='URL', action=URLAction, help='URL of an OCSP responder.')



# Generated by Django 1.10.5 on 2017-03-04 14:34

from django.db import migrations


def migrate_revocation_reasons(apps, schema_editor):
    Certificate = apps.get_model('django_ca', 'Certificate')
    certs = Certificate.objects.exclude(revoked_reason__isnull=True)

    for cert in certs.exclude(revoked_reason__in=['', 'unspecified', 'superseded']):
        if cert.revoked_reason == 'keyCompromise':
            cert.revoked_reason = 'key_compromise'
        elif cert.revoked_reason == 'caCompromise':
            cert.revoked_reason = 'ca_compromise'
        elif cert.revoked_reason == 'affiliationChanged':
            cert.revoked_reason = 'affiliation_changed'
        elif cert.revoked_reason == 'cessationOfOperation':
            cert.revoked_reason = 'cessation_of_operation'
        elif cert.revoked_reason == 'certificateHold':
            cert.revoked_reason = 'certificate_hold'
        else:
            raise RuntimeError('Unknown revocation reason encountered: %s' % cert.revoked_reason)

        cert.save()


class Migration(migrations.Migration):

    dependencies = [
        ('django_ca', '0002_auto_20170304_1434'),
    ]

    operations = [
        migrations.RunPython(migrate_revocation_reasons),
    ]



# Generated by Django 3.0.2 on 2020-01-19 11:27

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_ca', '0017_auto_20200112_1657'),
    ]

    operations = [
        migrations.AddField(
            model_name='certificate',
            name='profile',
            field=models.CharField(blank=True, default='', help_text='Profile that was used to generate this certificate.', max_length=32),
        ),
    ]



# This file is part of django-ca (https://github.com/mathiasertl/django-ca).
#
# django-ca is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# django-ca is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with django-ca.  If not,
# see <http://www.gnu.org/licenses/>.

"""Django models for the django-ca application.

.. seealso:: https://docs.djangoproject.com/en/dev/topics/db/models/
"""

import base64
import binascii
import hashlib
import importlib
import itertools
import json
import logging
import random
import re
from datetime import datetime
from datetime import timedelta

import pytz

from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import dsa
from cryptography.hazmat.primitives.serialization import Encoding
from cryptography.hazmat.primitives.serialization import PrivateFormat
from cryptography.hazmat.primitives.serialization import PublicFormat
from cryptography.hazmat.primitives.serialization import load_pem_private_key
from cryptography.x509.oid import ExtensionOID

from django.conf import settings
from django.core.cache import cache
from django.core.exceptions import ValidationError
from django.core.files.base import ContentFile
from django.core.validators import URLValidator
from django.db import models
from django.urls import reverse
from django.utils import timezone
from django.utils.crypto import get_random_string
from django.utils.encoding import force_bytes
from django.utils.encoding import force_str
from django.utils.functional import cached_property
from django.utils.text import slugify
from django.utils.translation import gettext_lazy as _

from . import ca_settings
from .acme.constants import BASE64_URL_ALPHABET
from .acme.constants import IdentifierType
from .acme.constants import Status
from .constants import ReasonFlags
from .extensions import OID_TO_EXTENSION
from .extensions import AuthorityInformationAccess
from .extensions import AuthorityKeyIdentifier
from .extensions import BasicConstraints
from .extensions import CertificatePolicies
from .extensions import CRLDistributionPoints
from .extensions import ExtendedKeyUsage
from .extensions import FreshestCRL
from .extensions import InhibitAnyPolicy
from .extensions import IssuerAlternativeName
from .extensions import KeyUsage
from .extensions import NameConstraints
from .extensions import OCSPNoCheck
from .extensions import PolicyConstraints
from .extensions import PrecertificateSignedCertificateTimestamps
from .extensions import PrecertPoison
from .extensions import SubjectAlternativeName
from .extensions import SubjectKeyIdentifier
from .extensions import TLSFeature
from .extensions import get_extension_name
from .managers import AcmeAccountManager
from .managers import AcmeAuthorizationManager
from .managers import AcmeCertificateManager
from .managers import AcmeChallengeManager
from .managers import AcmeOrderManager
from .managers import CertificateAuthorityManager
from .managers import CertificateManager
from .querysets import AcmeAccountQuerySet
from .querysets import AcmeAuthorizationQuerySet
from .querysets import AcmeCertificateQuerySet
from .querysets import AcmeChallengeQuerySet
from .querysets import AcmeOrderQuerySet
from .querysets import CertificateAuthorityQuerySet
from .querysets import CertificateQuerySet
from .signals import post_revoke_cert
from .signals import pre_revoke_cert
from .subject import Subject
from .utils import add_colons
from .utils import ca_storage
from .utils import format_name
from .utils import generate_private_key
from .utils import get_crl_cache_key
from .utils import int_to_hex
from .utils import multiline_url_validator
from .utils import parse_encoding
from .utils import parse_general_name
from .utils import parse_hash_algorithm
from .utils import read_file
from .utils import validate_key_parameters

log = logging.getLogger(__name__)


def acme_slug():
    """Default function to get an ACME conforming slug."""
    return get_random_string(length=12)


def acme_order_expires():
    """Default function for the expiry of an ACME order."""
    return timezone.now() + ca_settings.ACME_ORDER_VALIDITY


def acme_token():
    """Generate an ACME token for this challenge.

    Note that currently all challenges have the same requirements on tokens, except for DNS challenges
    which seem to allow padding ("=") characters. We ignore the '=' for DNS challenges as our tokens are
    already longer then required.
    """
    return get_random_string(64, allowed_chars=BASE64_URL_ALPHABET)


def validate_past(value):
    """Validate that a given datetime is not in the future."""
    if value > timezone.now():
        raise ValidationError(_('Date must be in the past!'))


def json_validator(value):
    """Validated that the given data is valid JSON."""
    try:
        json.loads(value)
    except Exception as e:
        raise ValidationError(_('Must be valid JSON: %(message)s') % {'message': str(e)}) from e


def pem_validator(value):
    """Validator that ensures a value is a valid PEM public certificate."""

    if not value.startswith('-----BEGIN PUBLIC KEY-----\n'):
        raise ValidationError(_('Not a valid PEM.'))
    if not value.endswith('\n-----END PUBLIC KEY-----'):
        raise ValidationError(_('Not a valid PEM.'))

    # TODO: for some reason cryptography cannot load LE account PEMs
    #try:
    #    x509.load_pem_x509_certificate(force_bytes(value), default_backend())
    #except Exception as ex:  # pylint: disable=broad-except
    #    raise ValidationError(_('Not a valid PEM.')) from ex


class DjangoCAModelMixin:
    """Mixin with shared properties for all django-ca models."""

    @property
    def admin_change_url(self):
        """Change URL in the admin interface for the given class."""
        return reverse('admin:%s_%s_change' % (self._meta.app_label, self._meta.model_name), args=(self.pk, ))


class Watcher(models.Model):
    """A watcher represents an email address that will receive notifications about expiring certificates."""
    name = models.CharField(max_length=64, blank=True, default='', verbose_name=_('CommonName'))
    mail = models.EmailField(verbose_name=_('E-Mail'), unique=True)

    @classmethod
    def from_addr(cls, addr):
        """Class constructor that creates an instance from an email address."""
        name = ''
        match = re.match(r'(.*?)\s*<(.*)>', addr)
        if match is not None:
            name, addr = match.groups()

        try:
            watcher = cls.objects.get(mail=addr)
            if watcher.name != name:
                watcher.name = name
                watcher.save()
        except cls.DoesNotExist:
            watcher = cls(mail=addr, name=name)
            watcher.full_clean()
            watcher.save()

        return watcher

    def __str__(self):
        if self.name:
            return '%s <%s>' % (self.name, self.mail)
        return self.mail


class X509CertMixin(DjangoCAModelMixin, models.Model):
    """Mixin class with common attributes for Certificates and Certificate Authorities."""
    # pylint: disable=too-many-instance-attributes,too-many-public-methods
    # X.509 certificates are complex. Sorry.

    # reasons are defined in http://www.ietf.org/rfc/rfc3280.txt
    REVOCATION_REASONS = (
        (ReasonFlags.aa_compromise.name, _('Attribute Authority compromised')),
        (ReasonFlags.affiliation_changed.name, _('Affiliation changed')),
        (ReasonFlags.ca_compromise.name, _('CA compromised')),
        (ReasonFlags.certificate_hold.name, _('On Hold')),
        (ReasonFlags.cessation_of_operation.name, _('Cessation of operation')),
        (ReasonFlags.key_compromise.name, _('Key compromised')),
        (ReasonFlags.privilege_withdrawn.name, _('Privilege withdrawn')),
        (ReasonFlags.remove_from_crl.name, _('Removed from CRL')),
        (ReasonFlags.superseded.name, _('Superseded')),
        (ReasonFlags.unspecified.name, _('Unspecified')),
    )

    created = models.DateTimeField(auto_now=True)

    valid_from = models.DateTimeField(blank=False)
    expires = models.DateTimeField(null=False, blank=False)

    pub = models.TextField(verbose_name=_('Public key'))
    cn = models.CharField(max_length=128, verbose_name=_('CommonName'))
    serial = models.CharField(max_length=64, unique=True)

    # revocation information
    revoked = models.BooleanField(default=False)
    revoked_date = models.DateTimeField(null=True, blank=True, verbose_name=_('Revoked on'),
                                        validators=[validate_past])
    revoked_reason = models.CharField(
        max_length=32, blank=True, default='', verbose_name=_('Reason for revokation'),
        choices=REVOCATION_REASONS)
    compromised = models.DateTimeField(
        null=True, blank=True, verbose_name=_('Date of compromise'), validators=[validate_past],
        help_text=_('Optional: When this certificate was compromised. You can change this date later.'))

    _x509 = None

    class Meta:
        abstract = True

    def get_revocation_reason(self):
        """Get the revocation reason of this certificate."""
        if self.revoked is False:
            return None

        return x509.ReasonFlags[self.revoked_reason]

    def get_compromised_time(self):
        """Return when this certificate was compromised as a *naive* datetime.

        Returns ``None`` if the time is not known **or** if the certificate is not revoked.
        """
        if self.revoked is False or not self.compromised:
            return None

        if timezone.is_aware(self.compromised):
            # convert datetime object to UTC and make it naive
            return timezone.make_naive(self.compromised, pytz.utc)

        return self.compromised

    def get_revocation_time(self):
        """Get the revocation time as naive datetime."""
        if self.revoked is False:
            return None

        if timezone.is_aware(self.revoked_date):
            # convert datetime object to UTC and make it naive
            return timezone.make_naive(self.revoked_date, pytz.utc)

        return self.revoked_date

    @property
    def x509(self):
        """The underlying :py:class:`cg:cryptography.x509.Certificate`."""
        if self._x509 is None:
            backend = default_backend()
            self._x509 = x509.load_pem_x509_certificate(force_bytes(self.pub), backend)
        return self._x509

    @x509.setter
    def x509(self, value):
        """Setter for the underlying :py:class:`cg:cryptography.x509.Certificate`."""
        self._x509 = value
        self.pub = force_str(self.dump_certificate(Encoding.PEM))
        self.cn = self.subject.get('CN', '')  # pylint: disable=invalid-name
        self.expires = self.not_after
        self.valid_from = self.not_before
        if settings.USE_TZ:
            self.expires = timezone.make_aware(self.expires, timezone=pytz.utc)
            self.valid_from = timezone.make_aware(self.valid_from, timezone=pytz.utc)

        self.serial = int_to_hex(value.serial_number)

    ##########################
    # Certificate properties #
    ##########################

    @property
    def algorithm(self):
        """A shortcut for :py:attr:`~cg:cryptography.x509.Certificate.signature_hash_algorithm`."""
        return self.x509.signature_hash_algorithm

    def dump_certificate(self, encoding=Encoding.PEM):
        """Get the certificate as bytes in the requested format.

        Parameters
        ----------

        encoding : attr of :py:class:`~cg:cryptography.hazmat.primitives.serialization.Encoding`, optional
            The format to return, defaults to ``Encoding.PEM``.
        """

        return self.x509.public_bytes(encoding=encoding)

    def get_digest(self, algo):
        """Get the digest for a certificate as string, including colons."""
        algo = getattr(hashes, algo.upper())()
        return add_colons(binascii.hexlify(self.x509.fingerprint(algo)).upper().decode('utf-8'))

    def get_filename(self, ext, bundle=False):
        """Get a filename safe for any file system and OS for this certificate based on the common name.

        Parameters
        ----------

        ext : str
            The filename extension to use (e.g. 'pem').
        bundle : bool, optional
            Adds "_bundle" as suffix.
        """
        slug = slugify(self.cn.replace('.', '_'))

        if bundle is True:
            return '%s_bundle.%s' % (slug, ext.lower())
        return '%s.%s' % (slug, ext.lower())

    def get_revocation(self):
        """Get the `RevokedCertificate` instance for this certificate for CRLs.

        This function is just a shortcut for
        :py:class:`~cg:cryptography.x509.RevokedCertificateBuilder`.

        .. seealso:: :py:class:`~cg:cryptography.x509.CertificateRevocationListBuilder`.

        Raises
        ------

        ValueError
            If the certificate is not revoked.

        Returns
        -------

        :py:class:`~cg:cryptography.x509.RevokedCertificate`
        """
        if self.revoked is False:
            raise ValueError('Certificate is not revoked.')

        revoked_cert = x509.RevokedCertificateBuilder().serial_number(
            self.x509.serial_number).revocation_date(self.revoked_date)

        reason = self.get_revocation_reason()
        if reason != x509.ReasonFlags.unspecified:
            # RFC 5270, 5.3.1: "reason code CRL entry extension SHOULD be absent instead of using the
            # unspecified (0) reasonCode value"
            revoked_cert = revoked_cert.add_extension(x509.CRLReason(reason), critical=False)

        compromised = self.get_compromised_time()
        if compromised:
            # RFC 5280, 5.3.2 says that this extension MUST be non-critical
            revoked_cert = revoked_cert.add_extension(x509.InvalidityDate(compromised), critical=False)

        return revoked_cert.build(default_backend())

    @property
    def hpkp_pin(self):
        """The HPKP public key pin for this certificate.

        Inspired by https://github.com/luisgf/hpkp-python/blob/master/hpkp.py.

        .. seealso:: https://en.wikipedia.org/wiki/HTTP_Public_Key_Pinning
        """

        public_key_raw = self.x509.public_key().public_bytes(
            encoding=Encoding.DER, format=PublicFormat.SubjectPublicKeyInfo)
        public_key_hash = hashlib.sha256(public_key_raw).digest()
        return base64.b64encode(public_key_hash).decode('utf-8')

    @property
    def issuer(self):
        """The certificate issuer field as :py:class:`~django_ca.subject.Subject`."""
        return Subject([(s.oid, s.value) for s in self.x509.issuer])

    @property
    def not_before(self):
        """Date/Time this certificate was created"""
        return self.x509.not_valid_before

    @property
    def not_after(self):
        """Date/Time this certificate expires."""
        return self.x509.not_valid_after

    def revoke(self, reason='', compromised=None):
        """Revoke the current certificate.

        This function emits the ``pre_revoke_cert`` and ``post_revoke_cert`` signals.

        Parameters
        ----------

        reason : :py:class:`~django_ca.constants.ReasonFlags`, optional
            The reason for revocation, defaults to ``ReasonFlags.unspecified``.
        compromised : datetime, optional
            When this certificate was compromised.
        """
        if not reason:
            reason = ReasonFlags.unspecified

        pre_revoke_cert.send(sender=self.__class__, cert=self, reason=reason)

        self.revoked = True
        self.revoked_date = timezone.now()
        self.revoked_reason = reason.name
        self.compromised = compromised
        self.save()

        post_revoke_cert.send(sender=self.__class__, cert=self)

    @property
    def subject(self):
        """The certificates subject as :py:class:`~django_ca.subject.Subject`."""
        return Subject([(s.oid, s.value) for s in self.x509.subject])

    @property
    def distinguished_name(self):
        """The certificates distinguished name formatted as string."""
        return format_name(self.x509.subject)
    distinguished_name.fget.short_description = 'Distinguished Name'

    ###################
    # X509 extensions #
    ###################

    @cached_property
    def _x509_extensions(self):
        return {e.oid: e for e in self.x509.extensions}

    def get_x509_extension(self, oid):
        """Get extension by a cryptography OID."""
        return self._x509_extensions.get(oid)

    @cached_property
    def _sorted_extensions(self):
        # NOTE: We need the dotted_string in the sort key if we have multiple unknown extensions, which then
        #       show up as "Unknown OID" and have to be sorted by oid
        return list(sorted(self._x509_extensions.values(),
                           key=lambda e: (get_extension_name(e), e.oid.dotted_string)))

    @cached_property
    def extension_fields(self):
        """List of all extensions fields for this certificate."""
        fields = []

        for ext in self._sorted_extensions:
            if ext.oid in OID_TO_EXTENSION:
                fields.append(OID_TO_EXTENSION[ext.oid].key)

            # extension that does not support new extension framework
            else:
                log.warning('Unknown extension encountered: %s (%s)',
                            get_extension_name(ext), ext.oid.dotted_string)
                fields.append(ext)
        return fields

    @cached_property
    def extensions(self):
        """List of all extensions for this certificate."""
        exts = []

        for ext in self._sorted_extensions:
            if ext.oid in OID_TO_EXTENSION:
                exts.append(getattr(self, OID_TO_EXTENSION[ext.oid].key))

            # extension that does not support new extension framework
            else:
                exts.append(ext)
        return exts

    @cached_property
    def authority_information_access(self):  # pylint: disable=inconsistent-return-statements
        """The :py:class:`~django_ca.extensions.AuthorityInformationAccess` extension or ``None`` if not
        present."""
        ext = self.get_x509_extension(ExtensionOID.AUTHORITY_INFORMATION_ACCESS)
        if ext is not None:
            return AuthorityInformationAccess(ext)

    @cached_property
    def authority_key_identifier(self):  # pylint: disable=inconsistent-return-statements
        """The :py:class:`~django_ca.extensions.AuthorityKeyIdentifier` extension or ``None`` if not
        present."""
        ext = self.get_x509_extension(ExtensionOID.AUTHORITY_KEY_IDENTIFIER)
        if ext is not None:
            return AuthorityKeyIdentifier(ext)

    @cached_property
    def basic_constraints(self):  # pylint: disable=inconsistent-return-statements
        """The :py:class:`~django_ca.extensions.BasicConstraints` extension or ``None`` if not present."""
        ext = self.get_x509_extension(ExtensionOID.BASIC_CONSTRAINTS)
        if ext is not None:
            return BasicConstraints(ext)

    @cached_property
    def crl_distribution_points(self):  # pylint: disable=inconsistent-return-statements
        """The :py:class:`~django_ca.extensions.CRLDistributionPoints` extension or ``None`` if not
        present."""
        ext = self.get_x509_extension(ExtensionOID.CRL_DISTRIBUTION_POINTS)
        if ext is not None:
            return CRLDistributionPoints(ext)

    @cached_property
    def certificate_policies(self):  # pylint: disable=inconsistent-return-statements
        """The :py:class:`~django_ca.extensions.CertificatePolicies` extension or ``None`` if not present."""
        ext = self.get_x509_extension(ExtensionOID.CERTIFICATE_POLICIES)
        if ext is not None:
            return CertificatePolicies(ext)

    @cached_property
    def freshest_crl(self):  # pylint: disable=inconsistent-return-statements
        """The :py:class:`~django_ca.extensions.FreshestCRL` extension or ``None`` if not present."""
        ext = self.get_x509_extension(ExtensionOID.FRESHEST_CRL)
        if ext is not None:
            return FreshestCRL(ext)

    @cached_property
    def inhibit_any_policy(self):  # pylint: disable=inconsistent-return-statements
        """The :py:class:`~django_ca.extensions.InhibitAnyPolicy` extension or ``None`` if not present."""
        ext = self.get_x509_extension(ExtensionOID.INHIBIT_ANY_POLICY)
        if ext is not None:
            return InhibitAnyPolicy(ext)

    @cached_property
    def issuer_alternative_name(self):  # pylint: disable=inconsistent-return-statements
        """The :py:class:`~django_ca.extensions.IssuerAlternativeName` extension or ``None`` if not
        present."""
        ext = self.get_x509_extension(ExtensionOID.ISSUER_ALTERNATIVE_NAME)
        if ext is not None:
            return IssuerAlternativeName(ext)

    @cached_property
    def policy_constraints(self):  # pylint: disable=inconsistent-return-statements
        """The :py:class:`~django_ca.extensions.PolicyConstraints` extension or ``None`` if not present."""
        ext = self.get_x509_extension(ExtensionOID.POLICY_CONSTRAINTS)
        if ext is not None:
            return PolicyConstraints(ext)

    @cached_property
    def key_usage(self):  # pylint: disable=inconsistent-return-statements
        """The :py:class:`~django_ca.extensions.KeyUsage` extension or ``None`` if not present."""
        ext = self.get_x509_extension(ExtensionOID.KEY_USAGE)
        if ext is not None:
            return KeyUsage(ext)

    @cached_property
    def extended_key_usage(self):  # pylint: disable=inconsistent-return-statements
        """The :py:class:`~django_ca.extensions.ExtendedKeyUsage` extension or ``None`` if not present."""
        ext = self.get_x509_extension(ExtensionOID.EXTENDED_KEY_USAGE)
        if ext is not None:
            return ExtendedKeyUsage(ext)

    @cached_property
    def name_constraints(self):  # pylint: disable=inconsistent-return-statements
        """The :py:class:`~django_ca.extensions.NameConstraints` extension or ``None`` if not present."""
        ext = self.get_x509_extension(ExtensionOID.NAME_CONSTRAINTS)
        if ext is not None:
            return NameConstraints(ext)

    @cached_property
    def ocsp_no_check(self):  # pylint: disable=inconsistent-return-statements
        """The :py:class:`~django_ca.extensions.OCSPNoCheck` extension or ``None`` if not present."""
        ext = self.get_x509_extension(ExtensionOID.OCSP_NO_CHECK)
        if ext is not None:
            return OCSPNoCheck(ext)

    @cached_property
    def precert_poison(self):  # pylint: disable=inconsistent-return-statements
        """The :py:class:`~django_ca.extensions.PrecertPoison` extension or ``None`` if not present."""
        ext = self.get_x509_extension(ExtensionOID.PRECERT_POISON)
        if ext is not None:
            return PrecertPoison(ext)

    @cached_property
    def precertificate_signed_certificate_timestamps(self):  # pylint: disable=inconsistent-return-statements
        """The :py:class:`~django_ca.extensions.PrecertificateSignedCertificateTimestamps` extension or
        ``None`` if not present."""
        ext = self.get_x509_extension(ExtensionOID.PRECERT_SIGNED_CERTIFICATE_TIMESTAMPS)
        if ext is not None:
            return PrecertificateSignedCertificateTimestamps(ext)

    @cached_property
    def subject_alternative_name(self):  # pylint: disable=inconsistent-return-statements
        """The :py:class:`~django_ca.extensions.SubjectAlternativeName` extension or ``None`` if not
        present."""
        ext = self.get_x509_extension(ExtensionOID.SUBJECT_ALTERNATIVE_NAME)
        if ext is not None:
            return SubjectAlternativeName(ext)

    @cached_property
    def subject_key_identifier(self):  # pylint: disable=inconsistent-return-statements
        """The :py:class:`~django_ca.extensions.SubjectKeyIdentifier` extension or ``None`` if not present."""
        ext = self.get_x509_extension(ExtensionOID.SUBJECT_KEY_IDENTIFIER)
        if ext is not None:
            return SubjectKeyIdentifier(ext)

    @cached_property
    def tls_feature(self):  # pylint: disable=inconsistent-return-statements
        """The :py:class:`~django_ca.extensions.TLSFeature` extension or ``None`` if not present."""
        ext = self.get_x509_extension(ExtensionOID.TLS_FEATURE)
        if ext is not None:
            return TLSFeature(ext)


class CertificateAuthority(X509CertMixin):
    """Model representing a x509 Certificate Authority."""

    objects = CertificateAuthorityManager.from_queryset(CertificateAuthorityQuerySet)()

    name = models.CharField(max_length=32, help_text=_('A human-readable name'), unique=True)
    """Human-readable name of the CA, only used for displaying the CA."""
    enabled = models.BooleanField(default=True)
    parent = models.ForeignKey('self', on_delete=models.SET_NULL, null=True, blank=True,
                               related_name='children')
    private_key_path = models.CharField(max_length=256, help_text=_('Path to the private key.'))

    # various details used when signing certs
    crl_url = models.TextField(blank=True, default='', validators=[multiline_url_validator],
                               verbose_name=_('CRL URLs'),
                               help_text=_("URLs, one per line, where you can retrieve the CRL."))
    crl_number = models.TextField(
        default='{"scope": {}}', blank=True, verbose_name=_('CRL Number'), validators=[json_validator],
        help_text=_("Data structure to store the CRL number (see RFC 5280, 5.2.3) depending on the scope.")
    )
    issuer_url = models.URLField(blank=True, null=True, verbose_name=_('Issuer URL'),
                                 help_text=_("URL to the certificate of this CA (in DER format)."))
    ocsp_url = models.URLField(blank=True, null=True, verbose_name=_('OCSP responder URL'),
                               help_text=_("URL of a OCSP responser for the CA."))
    issuer_alt_name = models.CharField(blank=True, max_length=255, default='',
                                       verbose_name=_('issuerAltName'), help_text=_("URL for your CA."))

    caa_identity = models.CharField(
        blank=True, max_length=32, verbose_name=_('CAA identity'),
        help_text=_('CAA identity for this CA (NOTE: Not currently used!).')
    )
    website = models.URLField(blank=True, help_text=_('Website for your CA.'))
    terms_of_service = models.URLField(blank=True, verbose_name='Terms of Service',
                                       help_text=_('URL to Terms of Service for this CA'))

    # ACMEv2 fields
    acme_enabled = models.BooleanField(
        default=False, verbose_name=_('Enable ACME'),
        help_text=_("Whether it is possible to use ACME for this CA."))
    acme_requires_contact = models.BooleanField(default=True, verbose_name='Requires contact', help_text=_(
        'If this CA requires a contact address during account registration.'))
    # CAA record and website are general fields

    _key = None

    def key(self, password):
        """The CAs private key as private key.

        .. seealso:: :py:func:`~cg:cryptography.hazmat.primitives.serialization.load_pem_private_key`.
        """
        if self._key is None:
            key_data = read_file(self.private_key_path)

            self._key = load_pem_private_key(key_data, password, default_backend())
        return self._key

    @property
    def key_exists(self):
        """``True`` if the private key is locally accessible."""
        if self._key is not None:
            return True
        return ca_storage.exists(self.private_key_path)

    def cache_crls(self, password=None, algorithm=None):  # pylint: disable=too-many-locals
        """Function to cache all CRLs for this CA."""

        password = password or self.get_password()
        ca_key = self.key(password)
        if isinstance(ca_key, dsa.DSAPrivateKey) and algorithm is None:
            algorithm = hashes.SHA1()

        for config in ca_settings.CA_CRL_PROFILES.values():
            overrides = config.get('OVERRIDES', {}).get(self.serial, {})

            if overrides.get('skip'):
                continue

            algorithm = algorithm or parse_hash_algorithm(overrides.get('algorithm', config.get('algorithm')))
            expires = overrides.get('expires', config.get('expires', 86400))
            scope = overrides.get('scope', config.get('scope'))
            full_name = overrides.get('full_name', config.get('full_name'))
            relative_name = overrides.get('relative_name', config.get('relative_name'))
            encodings = overrides.get('encodings', config.get('encodings', ['DER', ]))
            crl = None  # only compute crl when it is actually needed

            for encoding in encodings:
                encoding = parse_encoding(encoding)
                cache_key = get_crl_cache_key(self.serial, algorithm, encoding, scope=scope)

                if expires >= 600:  # pragma: no branch
                    # for longer expiries we substract a random value so that regular CRL regeneration is
                    # distributed a bit
                    cache_expires = expires - random.randint(1, 5) * 60

                if cache.get(cache_key) is None:
                    if crl is None:
                        crl = self.get_crl(expires=expires, algorithm=algorithm, password=password,
                                           scope=scope, full_name=full_name, relative_name=relative_name)

                    encoded_crl = crl.public_bytes(encoding)
                    cache.set(cache_key, encoded_crl, cache_expires)

    def generate_ocsp_key(self, profile='ocsp', expires=3, algorithm=None, password=None,
                          key_size=None, key_type=None, ecc_curve=None, autogenerated=True):
        """Generate OCSP keys for this CA.

        Parameters
        ----------

        profile : str, optional
            The profile to use for generating the certificate. The default is ``"ocsp"``.
        expires : int or datetime, optional
            Number of days or datetime when this certificate expires. The default is ``3`` (OCSP certificates
            are usually renewed frequently).
        algorithm : str, optional
            Passed to :py:func:`~django_ca.utils.parse_hash_algorithm` and defaults to
            :ref:`CA_DIGEST_ALGORITHM <settings-ca-digest-algorithm>`.
        password : bytes, optional
            The password to the CA as bytes, if its private key is encrypted.
        key_size : int, optional
            The key size of the private key, defaults to :ref:`CA_DEFAULT_KEY_SIZE
            <settings-ca-default-key-size>`.
        key_type : {"RSA", "DSA", "ECC"}, optional
            The private key type to use, the default is ``"RSA"``.
        ecc_curve : str, optional
            Passed to :py:func:`~django_ca.utils.parse_key_curve`, defaults to the :ref:`CA_DEFAULT_ECC_CURVE
            <settings-ca-default-ecc-curve>`.
        autogenerated : bool, optional
            Set the "autogenerated" flag of the certificate. ``True`` by default, since this method is usually
            invoked in an automated cron-like fashion.
        """
        # pylint: disable=too-many-arguments,too-many-locals
        # OCSP is pretty complex, there is no way to trim down the arguments w/o losing features.

        password = password or self.get_password()
        if key_type is None:
            ca_key = self.key(password)
            if isinstance(ca_key, dsa.DSAPrivateKey):
                key_type = 'DSA'
                algorithm = 'SHA1'

        key_size, key_type, ecc_curve = validate_key_parameters(key_size, key_type, ecc_curve)
        if isinstance(expires, int):
            expires = timedelta(days=expires)
        algorithm = parse_hash_algorithm(algorithm)

        # generate the private key
        private_key = generate_private_key(key_size, key_type, ecc_curve)
        private_pem = private_key.private_bytes(encoding=Encoding.PEM, format=PrivateFormat.PKCS8,
                                                encryption_algorithm=serialization.NoEncryption())
        private_path = ca_storage.generate_filename('ocsp/%s.key' % self.serial.replace(':', ''))

        csr = x509.CertificateSigningRequestBuilder().subject_name(self.x509.subject).sign(
            private_key, hashes.SHA256(), default_backend())

        # TODO: The subject we pass is just a guess - see what public CAs do!?  pylint: disable=fixme
        cert = Certificate.objects.create_cert(ca=self, csr=csr, profile=profile, subject=self.subject,
                                               algorithm=algorithm, autogenerated=autogenerated,
                                               password=password, add_ocsp_url=False)

        cert_path = ca_storage.generate_filename('ocsp/%s.pem' % self.serial.replace(':', ''))
        cert_pem = cert.dump_certificate(encoding=Encoding.PEM)

        for path, contents in [(private_path, private_pem), (cert_path, cert_pem)]:
            if ca_storage.exists(path):
                with ca_storage.open(path, 'wb') as stream:
                    stream.write(contents)
            else:
                ca_storage.save(path, ContentFile(contents))
        return private_path, cert_path, cert

    def get_authority_key_identifier(self):
        """Return the AuthorityKeyIdentifier extension used in certificates signed by this CA.

        Returns
        -------

        :py:class:`~cg:cryptography.x509.AuthorityKeyIdentifier`
            The value to use for this extension.
        """

        try:
            ski = self.x509.extensions.get_extension_for_class(x509.SubjectKeyIdentifier)
        except x509.ExtensionNotFound:
            return x509.AuthorityKeyIdentifier.from_issuer_public_key(self.x509.public_key())
        else:
            return x509.AuthorityKeyIdentifier.from_issuer_subject_key_identifier(ski.value)

    def get_authority_key_identifier_extension(self):
        """Get the AuthorityKeyIdentifier extension to use in certificates signed by this CA.

        Returns
        -------

        :py:class:`~django_ca.extensions.AuthorityKeyIdentifier`
            The extension to use.
        """

        return AuthorityKeyIdentifier(x509.Extension(
            critical=AuthorityKeyIdentifier.default_critical,
            oid=AuthorityKeyIdentifier.oid,
            value=self.get_authority_key_identifier()
        ))

    def get_crl(self, expires=86400, algorithm=None, password=None, scope=None, counter=None, **kwargs):
        """Generate a Certificate Revocation List (CRL).

        The ``full_name`` and ``relative_name`` parameters describe how to retrieve the CRL and are used in
        the `Issuing Distribution Point extension <https://tools.ietf.org/html/rfc5280.html#section-5.2.5>`_.
        The former defaults to the ``crl_url`` field, pass ``None`` to not include the value. At most one of
        the two may be set.

        Parameters
        ----------

        expires : int
            The time in seconds when this CRL expires. Note that you should generate a new CRL until then.
        algorithm : :py:class:`~cg:cryptography.hazmat.primitives.hashes.Hash` or str, optional
            The hash algorithm to use, passed to :py:func:`~django_ca.utils.parse_hash_algorithm`. The default
            is to use :ref:`CA_DIGEST_ALGORITHM <settings-ca-digest-algorithm>`.
        password : bytes, optional
            Password used to load the private key of the certificate authority. If not passed, the private key
            is assumed to be unencrypted.
        scope : {None, 'ca', 'user', 'attribute'}, optional
            What to include in the CRL: Use ``"ca"`` to include only revoked certificate authorities and
            ``"user"`` to include only certificates or ``None`` (the default) to include both.
            ``"attribute"`` is reserved for future use and always produces an empty CRL.
        counter : str, optional
            Override the counter-variable for the CRL Number extension. Passing the same key to multiple
            invocations will yield a different sequence then what would ordinarily be returned. The default is
            to use the scope as the key.
        full_name : list of str or :py:class:`~cg:cryptography.x509.GeneralName`, optional
            List of general names to use in the Issuing Distribution Point extension. If not passed, use
            ``crl_url`` if set.
        relative_name : :py:class:`~cg:cryptography.x509.RelativeDistinguishedName`, optional
            Used in Issuing Distribution Point extension, retrieve the CRL relative to the issuer.

        Returns
        -------

        bytes
            The CRL in the requested format.
        """
        # pylint: disable=too-many-statements,too-many-branches,too-many-locals,too-many-arguments
        # It's not easy to create a CRL. Sorry.

        if scope is not None and scope not in ['ca', 'user', 'attribute']:
            raise ValueError('Scope must be either None, "ca", "user" or "attribute"')

        now = now_builder = timezone.now()
        algorithm = parse_hash_algorithm(algorithm)

        if timezone.is_aware(now_builder):
            now_builder = timezone.make_naive(now, pytz.utc)
        else:
            now_builder = datetime.utcnow()

        builder = x509.CertificateRevocationListBuilder()
        builder = builder.issuer_name(self.x509.subject)
        builder = builder.last_update(now_builder)
        builder = builder.next_update(now_builder + timedelta(seconds=expires))

        if kwargs.get('full_name'):
            full_name = kwargs['full_name']
            full_name = [parse_general_name(n) for n in full_name]

        # CRLs for root CAs with scope "ca" (or no scope) do not add an IssuingDistributionPoint extension by
        # default. For full path validation with CRLs, the CRL is also used for validating the Root CA (which
        # does not contain a CRL Distribution Point). But the Full Name in the CRL IDP and the CA CRL DP have
        # to match. See also:
        #       https://github.com/mathiasertl/django-ca/issues/64
        elif scope in ('ca', None) and self.parent is None:
            full_name = None

        # If CA_DEFAULT_HOSTNAME is set, CRLs with scope "ca" add the same URL in the IssuingDistributionPoint
        # extension that is also added in the CRL Distribution Points extension for CAs issued by this CA.
        # See also:
        #       https://github.com/mathiasertl/django-ca/issues/64
        elif scope == 'ca' and ca_settings.CA_DEFAULT_HOSTNAME:
            crl_path = reverse('django_ca:ca-crl', kwargs={'serial': self.serial})
            full_name = [x509.UniformResourceIdentifier(
                'http://%s%s' % (ca_settings.CA_DEFAULT_HOSTNAME, crl_path)
            )]
        elif scope in ('user', None) and self.crl_url:
            crl_url = [url.strip() for url in self.crl_url.split()]
            full_name = [x509.UniformResourceIdentifier(c) for c in crl_url]
        else:
            full_name = None

        # Keyword arguments for the IssuingDistributionPoint extension
        idp_kwargs = {
            'only_contains_ca_certs': False,
            'only_contains_user_certs': False,
            'indirect_crl': False,
            'only_contains_attribute_certs': False,
            'only_some_reasons': None,
            'full_name': full_name,
            'relative_name': kwargs.get('relative_name'),
        }

        ca_qs = self.children.filter(expires__gt=now).revoked()
        cert_qs = self.certificate_set.filter(expires__gt=now).revoked()

        if scope == 'ca':
            certs = ca_qs
            idp_kwargs['only_contains_ca_certs'] = True
        elif scope == 'user':
            certs = cert_qs
            idp_kwargs['only_contains_user_certs'] = True
        elif scope == 'attribute':
            # sorry, nothing we support right now
            certs = []
            idp_kwargs['only_contains_attribute_certs'] = True
        else:
            certs = itertools.chain(ca_qs, cert_qs)

        for cert in certs:
            builder = builder.add_revoked_certificate(cert.get_revocation())

        # We can only add the IDP extension if one of these properties is set, see RFC 5280, 5.2.5.
        add_idp = idp_kwargs['only_contains_attribute_certs'] or idp_kwargs['only_contains_user_certs'] \
            or idp_kwargs['only_contains_ca_certs'] or idp_kwargs['full_name'] or idp_kwargs['relative_name']

        if add_idp:  # pragma: no branch
            builder = builder.add_extension(x509.IssuingDistributionPoint(**idp_kwargs), critical=True)

        # Add AuthorityKeyIdentifier from CA
        aki = self.get_authority_key_identifier()
        builder = builder.add_extension(aki, critical=False)

        # Add the CRLNumber extension (RFC 5280, 5.2.3)
        if counter is None:
            counter = scope or 'all'
        crl_number_data = json.loads(self.crl_number)
        crl_number = int(crl_number_data['scope'].get(counter, 0))
        builder = builder.add_extension(x509.CRLNumber(crl_number=crl_number), critical=False)

        # increase crl_number for the given scope and save
        crl_number_data['scope'][counter] = crl_number + 1
        self.crl_number = json.dumps(crl_number_data)
        self.save()

        return builder.sign(private_key=self.key(password), algorithm=algorithm, backend=default_backend())

    def get_password(self):
        """Get password for the private key from the ``CA_PASSWORDS`` setting."""
        return ca_settings.CA_PASSWORDS.get(self.serial)

    @property
    def pathlen(self):
        """The ``pathlen`` attribute of the ``BasicConstraints`` extension (either an ``int`` or ``None``)."""

        try:
            ext = self.x509.extensions.get_extension_for_oid(ExtensionOID.BASIC_CONSTRAINTS)
        except x509.ExtensionNotFound:  # pragma: no cover - extension should always be present
            return None
        return ext.value.path_length

    @property
    def max_pathlen(self):
        """The maximum pathlen for any intermediate CAs signed by this CA.

        This value is either ``None``, if this and all parent CAs don't have a ``pathlen`` attribute, or an
        ``int`` if any parent CA has the attribute.
        """

        pathlen = self.pathlen
        if self.parent is None:
            return pathlen

        max_parent = self.parent.max_pathlen

        if max_parent is None:
            return pathlen
        if pathlen is None:
            return max_parent - 1

        return min(self.pathlen, max_parent - 1)

    @property
    def allows_intermediate_ca(self):
        """Wether this CA allows creating intermediate CAs."""

        max_pathlen = self.max_pathlen
        return max_pathlen is None or max_pathlen > 0

    @property
    def bundle(self):
        """A list of any parent CAs, including this CA.

        The list is ordered so the Root CA will be the first.
        """
        ca = self
        bundle = [ca]

        while ca.parent is not None:
            bundle.append(ca.parent)
            ca = ca.parent
        return bundle

    @property
    def root(self):
        """Get the root CA for this CA."""

        if self.parent is None:
            return self

        ca = self
        while ca.parent is not None:
            ca = ca.parent
        return ca

    @property
    def usable(self):
        """True if the CA is currently usable or not."""
        return self.enabled and self.valid_from < timezone.now() < self.expires

    class Meta:
        verbose_name = _('Certificate Authority')
        verbose_name_plural = _('Certificate Authorities')

    def __str__(self):
        return self.name


class Certificate(X509CertMixin):
    """Model representing a x509 Certificate."""

    objects = CertificateManager.from_queryset(CertificateQuerySet)()

    watchers = models.ManyToManyField(Watcher, related_name='certificates', blank=True)

    ca = models.ForeignKey(CertificateAuthority, on_delete=models.CASCADE,
                           verbose_name=_('Certificate Authority'))
    csr = models.TextField(verbose_name=_('CSR'), blank=True)

    # Note: We don't set choices here because the available profiles might be changed by the user.
    profile = models.CharField(blank=True, default='', max_length=32,
                               help_text=_('Profile that was used to generate this certificate.'))

    autogenerated = models.BooleanField(default=False,
                                        help_text=_("If this certificate was automatically generated."))

    @property
    def bundle(self):
        """The complete certificate bundle. This includes all CAs as well as the certificates itself."""

        return [self] + self.ca.bundle

    @property
    def root(self):
        """Get the root CA for this certificate."""

        return self.ca.root

    def __str__(self):
        return self.cn


class AcmeAccount(DjangoCAModelMixin, models.Model):
    """Implements an ACME account object.

    .. seealso::

        `RFC 8555, 7.1.2 <https://tools.ietf.org/html/rfc8555#section-7.1.2>`_
    """

    # RFC 8555, 7.1.2: "Possible values are "valid", "deactivated", and "revoked"."
    STATUS_VALID = Status.VALID.value
    STATUS_DEACTIVATED = Status.DEACTIVATED.value  # deactivated by user
    STATUS_REVOKED = Status.REVOKED.value  # revoked by server
    STATUS_CHOICES = (
        (STATUS_VALID, _('Valid')),
        (STATUS_DEACTIVATED, _('Deactivated')),
        (STATUS_REVOKED, _('Revoked')),
    )

    objects = AcmeAccountManager.from_queryset(AcmeAccountQuerySet)()

    # Account meta data
    created = models.DateTimeField(default=timezone.now)
    updated = models.DateTimeField(auto_now=True)

    # Account information
    ca = models.ForeignKey(CertificateAuthority, on_delete=models.CASCADE,
                           verbose_name=_('Certificate Authority'))
    # Full public key of the account
    pem = models.TextField(verbose_name=_('Public key'), unique=True, blank=False, validators=[pem_validator])
    # JSON Web Key thumbprint - a hash of the public key, see RFC 7638.
    #   NOTE: Only unique for the given CA to make hash collisions less likely
    thumbprint = models.CharField(max_length=64)
    slug = models.SlugField(unique=True, default=acme_slug)
    kid = models.URLField(unique=True, validators=[URLValidator(schemes=('http', 'https'))],
                          verbose_name=_('Key ID'))

    # Fields according to RFC 8555, 7.1.2
    # RFC 8555, 7.1.6: "Account objects are created in the "valid" state"
    status = models.CharField(choices=STATUS_CHOICES, max_length=12, default=STATUS_VALID)
    contact = models.TextField(blank=True, help_text=_('Contact addresses for this account, one per line.'))
    terms_of_service_agreed = models.BooleanField(default=False)
    # NOTE: externalAccountBinding is not yet supported
    # NOTE: orders property is provided by reverse relation of the AcmeOrder model

    class Meta:
        verbose_name = _('ACME Account')
        verbose_name_plural = _('ACME Accounts')
        unique_together = (
            ('ca', 'thumbprint'),
        )

    def __str__(self):
        try:
            return self.contact.split('\n')[0].split(':', 1)[1]
        except IndexError:
            return ''

    @property
    def serial(self):
        """Serial of the CA for this account."""
        return self.ca.serial

    def set_kid(self, request):
        """Set the ACME kid based on this accounts CA and slug.

        Note that `slug` and `ca` must be already set when using this method.
        """
        self.kid = request.build_absolute_uri(
            reverse('django_ca:acme-account', kwargs={'slug': self.slug, 'serial': self.ca.serial})
        )

    @property
    def usable(self):
        """Boolean if the account is currently usable.

        An account is usable if the terms of service have been agreed, the status is "valid" and the
        associated CA is usable.
        """
        return self.terms_of_service_agreed and self.status == AcmeAccount.STATUS_VALID and self.ca.usable


class AcmeOrder(DjangoCAModelMixin, models.Model):
    """Implements an ACME order object.

    .. seealso::

        `RFC 8555, 7.1.3 <https://tools.ietf.org/html/rfc8555#section-7.1.3>`_
    """
    # RFC 8555, 7.1.3: "Possible values are "pending", "ready", "processing", "valid", and "invalid"."
    STATUS_PENDING = Status.PENDING.value
    STATUS_READY = Status.READY.value
    STATUS_PROCESSING = Status.PROCESSING.value
    STATUS_VALID = Status.VALID.value
    STATUS_INVALID = Status.INVALID.value

    STATUS_CHOICES = (
        (STATUS_INVALID, _('Invalid')),
        (STATUS_PENDING, _('Pending')),
        (STATUS_PROCESSING, _('Processing')),
        (STATUS_READY, _('Ready')),
        (STATUS_VALID, _('Valid')),
    )

    objects = AcmeOrderManager.from_queryset(AcmeOrderQuerySet)()

    account = models.ForeignKey(AcmeAccount, on_delete=models.CASCADE, related_name='orders')
    slug = models.SlugField(unique=True, default=acme_slug)

    # Fields according to RFC 8555, 7.1.3
    # RFC 8555, 7.1.6: "Order objects are created in the "pending" state."
    status = models.CharField(choices=STATUS_CHOICES, max_length=10, default=STATUS_PENDING)
    expires = models.DateTimeField(default=acme_order_expires)
    # NOTE: identifiers property is provided by reverse relation of the AcmeAuthorization model
    not_before = models.DateTimeField(null=True)
    not_after = models.DateTimeField(null=True)
    # NOTE: error property is not yet supported
    # NOTE: authorizations property is provided by reverse relation of the AcmeAuthorization model
    # NOTE: finalize property is provided by acme_finalize_url property
    # NOTE: certificate property is provided by reverse relation of the AcmeCertificate model

    class Meta:
        verbose_name = _('ACME Order')
        verbose_name_plural = _('ACME Orders')

    def __str__(self):
        return '%s (%s)' % (self.slug, self.account)

    @property
    def acme_url(self):
        """Get the ACME url path for this order."""
        return reverse('django_ca:acme-order', kwargs={'slug': self.slug, 'serial': self.serial})

    @property
    def acme_finalize_url(self):
        """Get the ACME "finalize" url path for this order."""
        return reverse('django_ca:acme-order-finalize', kwargs={'slug': self.slug, 'serial': self.serial})

    def add_authorizations(self, identifiers):
        """Add :py:class:`~django_ca.models.AcmeAuthorization` instances for the given identifiers.

        Note that this method already adds the account authorization to the database. It does not verify if it
        already exists and will raise an IntegrityError if it does.

        Example::

            >>> from acme import messages
            >>> identifier = messages.Identifier(typ=messages.IDENTIFIER_FQDN, value='example.com')
            >>> order.add_authorizations([identifier])

        Parameters
        ----------

        identifiers : list of :py:class:`acme:acme.messages.Identifier`
            The identifiers for this for this order.

        Returns
        -------

        list of :py:class:`~django_ca.models.AcmeAuthorization`
        """
        return self.authorizations.bulk_create([
            AcmeAuthorization(type=ident.typ.name, value=ident.value, order=self) for ident in identifiers
        ])

    @property
    def serial(self):
        """Serial of the CA for this order."""
        return self.account.serial

    @property
    def usable(self):
        """Boolean defining if an order is "usable", meaning it can be used to issue a certificate.

        An order is usable if it is in the "pending" status, has not expired and the account is usable.
        """
        return self.status == AcmeOrder.STATUS_PENDING and self.expires > timezone.now() \
            and self.account.usable


class AcmeAuthorization(models.Model):
    """Implements an ACME authorization object.

    .. seealso::

        `RFC 8555, 7.1.4 <https://tools.ietf.org/html/rfc8555#section-7.1.4>`_
    """
    # Choices from RFC 8555, section 9.7.7.
    TYPE_DNS = IdentifierType.DNS.value
    TYPE_CHOICES = (
        (TYPE_DNS, _('DNS')),
    )

    # RFC 8555, 7.1.4: "Possible values are "pending", "valid", "invalid", "deactivated", "expired", and
    #                   "revoked"."
    STATUS_PENDING = Status.PENDING.value
    STATUS_VALID = Status.VALID.value
    STATUS_INVALID = Status.INVALID.value
    STATUS_DEACTIVATED = Status.DEACTIVATED.value
    STATUS_EXPIRED = Status.EXPIRED.value
    STATUS_REVOKED = Status.REVOKED.value
    STATUS_CHOICES = (
        (STATUS_PENDING, _('Pending')),
        (STATUS_VALID, _('Valid')),
        (STATUS_INVALID, _('Invalid')),
        (STATUS_DEACTIVATED, _('Deactivated')),
        (STATUS_EXPIRED, _('Expired')),
        (STATUS_REVOKED, _('Revoked')),
    )

    objects = AcmeAuthorizationManager.from_queryset(AcmeAuthorizationQuerySet)()

    order = models.ForeignKey(AcmeOrder, on_delete=models.CASCADE, related_name='authorizations')
    slug = models.SlugField(unique=True, default=acme_slug)

    # Fields according to RFC 8555, 7.1.4:
    # NOTE: RFC 8555 does not specify a default value but DNS is the only known value
    type = models.CharField(choices=TYPE_CHOICES, max_length=8, default=TYPE_DNS)  # identifier
    value = models.CharField(max_length=255)  # identifier
    # RFC 8555, 7.1.6: "Authorization objects are created in the "pending" state."
    status = models.CharField(choices=STATUS_CHOICES, max_length=12, default=STATUS_PENDING)
    # NOTE: expires property comes from the linked order
    # NOTE: challenges property is provided by reverse relation of the AcmeChallenge model
    wildcard = models.BooleanField(default=False)

    class Meta:
        unique_together = (
            ('order', 'type', 'value'),
        )
        verbose_name = _('ACME Authorization')
        verbose_name_plural = _('ACME Authorizations')

    def __str__(self):
        return '%s: %s' % (self.type, self.value)

    @property
    def account(self):
        """Account that this authorization belongs to."""
        return self.order.account

    @property
    def acme_url(self):
        """Get the ACME url path for this account authorization."""
        return reverse('django_ca:acme-authz', kwargs={'slug': self.slug, 'serial': self.serial})

    @property
    def expires(self):
        """When this authorization expires."""
        return self.order.expires  # so far there is no reason to have a different value here

    @property
    def identifier(self):
        """Get ACME identifier for this object.

        Returns
        -------

        identifier : :py:class:`acme:acme.messages.Identifier`
        """
        # Programatic import to make sure that the acme library is an optional dependency
        messages = importlib.import_module('acme.messages')

        if self.type == AcmeAuthorization.TYPE_DNS:
            return messages.Identifier(typ=messages.IDENTIFIER_FQDN, value=self.value)
        raise ValueError('Unknown identifier type: %s' % self.type)

    @property
    def serial(self):
        """Serial of the CA for this authorization."""
        return self.order.serial

    @property
    def subject_alternative_name(self):
        """Get the domain for this challenge as prefixed SubjectAlternativeName.

        This method is intended to be used when creating the
        :py:class:`~django_ca.extensions.SubjectAlternativeName` extension for a certificate to be signed.
        """
        return '%s:%s' % (self.type, self.value)

    def get_challenges(self):
        """Get list of :py:class:`~django_ca.models.AcmeChallenge` objects for this authorization.

        Note that challenges will be created if they don't exist.
        """

        return [
            AcmeChallenge.objects.get_or_create(auth=self, type=AcmeChallenge.TYPE_HTTP_01)[0],
            AcmeChallenge.objects.get_or_create(auth=self, type=AcmeChallenge.TYPE_DNS_01)[0],
        ]

    @property
    def usable(self):
        """Boolean defining if an auth is "usable", meaning it still can be used in order validation

        An order is usable if it is in the "pending" or "invalid" status, the order is usable. An
        authorization that is in the "invalid" status is eligible to be retried by the client.
        """
        states = (AcmeAuthorization.STATUS_PENDING, AcmeAuthorization.STATUS_INVALID)
        return self.status in states and self.order.usable


class AcmeChallenge(models.Model):
    """Implements an ACME Challenge Object.

    .. seealso:: `RFC 8555, section 7.1.5 <https://tools.ietf.org/html/rfc8555#section-7.1.5>`_
    """

    # Possible challenges
    TYPE_HTTP_01 = 'http-01'
    TYPE_DNS_01 = 'dns-01'
    TYPE_TLS_ALPN_01 = 'tls-alpn-01'
    TYPE_CHOICES = (
        (TYPE_HTTP_01, _('HTTP Challenge')),
        (TYPE_DNS_01, _('DNS Challenge')),
        (TYPE_TLS_ALPN_01, _('TLS ALPN Challenge')),
    )

    # RFC 8555, 8: "Possible values are "pending", "processing", "valid", and "invalid"."
    STATUS_PENDING = Status.PENDING.value
    STATUS_PROCESSING = Status.PROCESSING.value
    STATUS_VALID = Status.VALID.value
    STATUS_INVALID = Status.INVALID.value
    STATUS_CHOICES = (
        (STATUS_PENDING, _('Pending')),
        (STATUS_PROCESSING, _('Processing')),
        (STATUS_VALID, _('Valid')),
        (STATUS_INVALID, _('Name')),
    )

    objects = AcmeChallengeManager.from_queryset(AcmeChallengeQuerySet)()

    auth = models.ForeignKey(AcmeAuthorization, on_delete=models.CASCADE, related_name='challenges')
    slug = models.SlugField(unique=True, default=acme_slug)

    # Fields according to RFC 8555, 8:
    type = models.CharField(choices=TYPE_CHOICES, max_length=12)
    # NOTE: url property is provided by the acme_url property and computed on the fly
    # RFC 8555, 7.1.6: "Challenge objects are created in the "pending" state."
    status = models.CharField(choices=STATUS_CHOICES, max_length=12, default=STATUS_PENDING)
    validated = models.DateTimeField(null=True, blank=True)
    error = models.CharField(blank=True, max_length=64)  # max_length is just a guess

    # The token field is listed for both HTTP and DNS challenge, which are the most common types, so we
    # include it as an optional field here. It is generated when the token is first accessed.
    token = models.CharField(blank=True, max_length=64, default=acme_token)

    class Meta:
        unique_together = (
            ('auth', 'type'),
        )
        verbose_name = _('ACME Challenge')
        verbose_name_plural = _('ACME Challenges')

    def __str__(self):
        return '%s (%s)' % (self.auth.value, self.type)

    @property
    def acme_url(self):
        """Get the ACME url path for this challenge."""
        return reverse('django_ca:acme-challenge', kwargs={'slug': self.slug, 'serial': self.serial})

    @property
    def acme_challenge(self):
        """Challenge as ACME challenge object.

        Returns
        -------

        acme.messages.Challenge
            The acme representation of this class.
        """
        # Programatic import to make sure that the acme library is an optional dependency
        challenges = importlib.import_module('acme.challenges')

        token = self.token.encode()
        if self.type == AcmeChallenge.TYPE_HTTP_01:
            return challenges.HTTP01(token=token)
        if self.type == AcmeChallenge.TYPE_DNS_01:
            return challenges.DNS01(token=token)
        if self.type == AcmeChallenge.TYPE_TLS_ALPN_01:
            return challenges.TLSALPN01(token=token)

        raise ValueError('%s: Unsupported challenge type.' % self.type)

    @property
    def acme_validated(self):
        """Timestamp when this challenge was validated.

        This property is a wrapper around the `validated` field. It always returns `None` if the challenge is
        not marked as valid (even if it had a timestamp), and the timestamp will always have a timezone, even
        if ``USE_TZ=False``.
        """
        if self.status != AcmeChallenge.STATUS_VALID or self.validated is None:
            return None

        if timezone.is_naive(self.validated):
            return timezone.make_aware(self.validated, timezone=pytz.UTC)
        return self.validated

    def get_challenge(self, request):
        """Get the ACME challenge body for this challenge.

        Returns
        -------

        acme.messages.ChallengeBody
            The acme representation of this class.
        """

        # Programatic import to make sure that the acme library is an optional dependency
        messages = importlib.import_module('acme.messages')

        url = request.build_absolute_uri(self.acme_url)

        # NOTE: RFC855, section 7.5 shows challenges *without* a status, but this object always includes it.
        #       It does not seem to hurt, but might be a slight spec-violation.
        return messages.ChallengeBody(chall=self.acme_challenge, _url=url, status=self.status,
                                      validated=self.acme_validated)

    @property
    def serial(self):
        """Serial of the CA for this challenge."""
        return self.auth.serial

    @property
    def usable(self):
        """Boolean defining if an challenge is "usable", meaning it still can be used in order validation.

        A challenge is usable if it is in the "pending" or "invalid status and the authorization is usable.
        """
        states = (AcmeChallenge.STATUS_PENDING, AcmeChallenge.STATUS_INVALID)
        return self.status in states and self.auth.usable


class AcmeCertificate(models.Model):
    """Intermediate model for certificates to be issued via ACME."""

    objects = AcmeCertificateManager.from_queryset(AcmeCertificateQuerySet)()

    slug = models.SlugField(unique=True, default=acme_slug)
    order = models.OneToOneField(AcmeOrder, on_delete=models.CASCADE)
    cert = models.OneToOneField(Certificate, on_delete=models.CASCADE, null=True)
    csr = models.TextField(verbose_name=_('CSR'))

    class Meta:
        verbose_name = _('ACME Certificate')
        verbose_name_plural = _('ACME Certificate')

    @property
    def acme_url(self):
        """Get the ACME url path for this certificate."""
        return reverse('django_ca:acme-cert', kwargs={'slug': self.slug, 'serial': self.order.serial})

    def parse_csr(self):
        """Load the CSR into a cryptography object.

        Returns
        -------

        :py:class:`~cg:cryptography.x509.CertificateSigningRequest`
            The CSR as used by cryptography.
        """
        return x509.load_pem_x509_csr(self.csr.encode(), default_backend())

    @property
    def usable(self):
        """Boolean defining if this instance is "usable", meaning we can use it to issue a certificate.

        An ACME certificate is considered usable if no actuall certificate has yet been issued, the order is
        not expired and in the "processing" state.
        """
        return self.cert is None and self.order.expires > timezone.now() \
            and self.order.status == AcmeOrder.STATUS_PROCESSING



# This file is part of django-ca (https://github.com/mathiasertl/django-ca).
#
# django-ca is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# django-ca is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with django-ca.  If not,
# see <http://www.gnu.org/licenses/>.

"""Management command to write a CRL to stdout or a file.

.. seealso:: https://docs.djangoproject.com/en/dev/howto/custom-management-commands/
"""

from django.core.management.base import CommandError

from ..base import BaseCommand


class Command(BaseCommand):  # pylint: disable=missing-class-docstring
    help = "Write the certificate revocation list (CRL)."
    binary_output = True

    def add_arguments(self, parser):
        parser.add_argument(
            '-e', '--expires', type=int, default=86400, metavar='SECONDS',
            help="Seconds until a new CRL will be available (default: %(default)s).")
        parser.add_argument('path', nargs='?', default='-',
                            help='Path for the output file. Use "-" for stdout.')
        parser.add_argument(
            '--ca-crl', action='store_true', default=False,
            help="*DEPRECATED:* Use --scope=ca instead. Generate the CRL for revoked child CAs.")
        parser.add_argument(
            '-s', '--scope', choices=['ca', 'user', 'attribute'],
            help='Limit the scope for the CRL (default: %(default)s).')
        self.add_algorithm(parser)
        self.add_format(parser)
        self.add_ca(parser, allow_disabled=True)
        self.add_password(parser)
        super().add_arguments(parser)

    def handle(self, path, **options):  # pylint: disable=arguments-differ
        if options['ca_crl']:
            self.stderr.write(self.style.WARNING('WARNING: --ca-crl is deprecated, use --scope=ca instead.'))
            options['scope'] = 'ca'

        kwargs = {
            'expires': options['expires'],
            'algorithm': options['algorithm'],
            'password': options['password'],
            'scope': options['scope'],
        }

        # See if we can work with the private key
        ca = options['ca']
        self.test_private_key(ca, options['password'])

        try:
            crl = ca.get_crl(**kwargs).public_bytes(options['format'])
        except Exception as ex:
            # Note: all parameters are already sanitized by parser actions
            raise CommandError(ex) from ex

        if path == '-':
            self.stdout.write(crl, ending=b'')
        else:
            try:
                with open(path, 'wb') as stream:
                    stream.write(crl)
            except IOError as ex:
                raise CommandError(ex) from ex



# This file is part of django-ca (https://github.com/mathiasertl/django-ca).
#
# django-ca is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# django-ca is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with django-ca.  If not,
# see <http://www.gnu.org/licenses/>.

from django.forms import widgets
from django.utils.encoding import force_str
from django.utils.translation import gettext as _

from . import ca_settings


class LabeledCheckboxInput(widgets.CheckboxInput):
    """CheckboxInput widget that adds a label and wraps everything in a <span />.

    This is necessary because widgets in MultiValueFields don't render with a label."""

    template_name = 'django_ca/forms/widgets/labeledcheckboxinput.html'

    def __init__(self, label, *args, **kwargs):
        self.label = label
        super(LabeledCheckboxInput, self).__init__(*args, **kwargs)

    def get_context(self, *args, **kwargs):
        ctx = super(LabeledCheckboxInput, self).get_context(*args, **kwargs)
        ctx['widget']['label'] = self.label
        return ctx

    class Media:
        css = {
            'all': ('django_ca/admin/css/labeledcheckboxinput.css', ),
        }


class LabeledTextInput(widgets.TextInput):
    """CheckboxInput widget that adds a label and wraps everything in a <span />.

    This is necessary because widgets in MultiValueFields don't render with a label."""

    template_name = 'django_ca/forms/widgets/labeledtextinput.html'

    def __init__(self, label, *args, **kwargs):
        self.label = label
        super(LabeledTextInput, self).__init__(*args, **kwargs)

    def get_context(self, *args, **kwargs):
        ctx = super(LabeledTextInput, self).get_context(*args, **kwargs)
        ctx['widget']['label'] = self.label
        ctx['widget']['cssid'] = self.label.lower().replace(' ', '-')
        return ctx

    class Media:
        css = {
            'all': ('django_ca/admin/css/labeledtextinput.css', ),
        }


class SubjectTextInput(LabeledTextInput):
    template_name = 'django_ca/forms/widgets/subjecttextinput.html'


class ProfileWidget(widgets.Select):
    # TODO: shouldn't we set a template_name here? Perhaps that's why render() is still called

    def render(self, name, value, attrs=None, renderer=None):
        html = super(ProfileWidget, self).render(name, value, attrs=attrs, renderer=renderer)

        # add the description of the default selected profile as help text (will be updated by JS when
        # different profile is selected)
        desc = ca_settings.CA_PROFILES[ca_settings.CA_DEFAULT_PROFILE].get(
            'description', ca_settings.CA_PROFILES[ca_settings.CA_DEFAULT_PROFILE].get('desc', ''))
        html += '<p class="help profile-desc">%s</p>' % force_str(desc)

        return html

    class Media:
        js = (
            'admin/js/jquery.init.js',
            'django_ca/admin/js/profilewidget.js',
        )


class CustomMultiWidget(widgets.MultiWidget):
    """Wraps the multi widget into a <p> element."""

    template_name = 'django_ca/forms/widgets/custommultiwidget.html'


class SubjectWidget(CustomMultiWidget):
    def __init__(self, attrs=None):
        _widgets = (
            SubjectTextInput(label=_('Country'), attrs={'placeholder': '2 character country code'}),
            SubjectTextInput(label=_('State')),
            SubjectTextInput(label=_('Location')),
            SubjectTextInput(label=_('Organization')),
            SubjectTextInput(label=_('Organizational Unit')),
            SubjectTextInput(label=_('CommonName'), attrs={'required': True}),
            SubjectTextInput(label=_('E-Mail')),
        )
        super(SubjectWidget, self).__init__(_widgets, attrs)

    def decompress(self, value):
        if value is None:  # pragma: no cover
            return ('', '', '', '', '', '')

        # Multiple OUs are not supported in webinterface
        ou = value.get('OU', '')
        if isinstance(ou, list) and ou:
            ou = ou[0]

        # Used e.g. for initial form data (e.g. resigning a cert)
        return [
            value.get('C', ''),
            value.get('ST', ''),
            value.get('L', ''),
            value.get('O', ''),
            ou,
            value.get('CN', ''),
            value.get('emailAddress', ''),
        ]


class SubjectAltNameWidget(CustomMultiWidget):
    def __init__(self, attrs=None):
        _widgets = (
            widgets.TextInput(),
            LabeledCheckboxInput(label="Include CommonName")
        )
        super(SubjectAltNameWidget, self).__init__(_widgets, attrs)

    def decompress(self, value):  # pragma: no cover
        if value:
            return value
        return ('', True)


class MultiValueExtensionWidget(CustomMultiWidget):
    def __init__(self, choices, attrs=None):
        _widgets = (
            widgets.SelectMultiple(choices=choices, attrs=attrs),
            LabeledCheckboxInput(label=_('critical')),
        )
        super(MultiValueExtensionWidget, self).__init__(_widgets, attrs)

    def decompress(self, value):
        if value:
            return value.serialize_iterable(), value.critical
        return ([], False)



# This file is part of django-ca (https://github.com/mathiasertl/django-ca).
#
# django-ca is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# django-ca is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with django-ca.  If not,
# see <http://www.gnu.org/licenses/>.

"""Management command to add/remove certificate watchers.

.. seealso:: https://docs.djangoproject.com/en/dev/howto/custom-management-commands/
"""

from django_ca.management.base import CertCommand
from django_ca.models import Watcher


class Command(CertCommand):  # pylint: disable=missing-class-docstring
    help = '''Add/remove addresses to be notified of an expiring certificate. The
        "list_certs" command lists all known certificates.

        E-Mail addresses can be verbatim ("user@example.com") or with a name, e.g. "Your Name
        <user@example.com>", the latter case must be quoted on the shell.
        '''

    def add_arguments(self, parser):
        parser.add_argument(
            '-a', '--add', metavar='EMAIL', default=[], action='append',
            help='''Address that now should be notified when the certificate expires. Add an email
                to be notified of an expiring certificate (may be given multiple times).''')
        parser.add_argument(
            '-r', '--rm', metavar='EMAIL', default=[], action='append',
            help='''Address that shoult no longer be notified when the certificate expires
                (may be given multiple times).''')
        super().add_arguments(parser)

    def handle(self, cert, **options):  # pylint: disable=arguments-differ
        # add/remove users:
        cert.watchers.add(*[Watcher.from_addr(addr) for addr in options['add']])
        cert.watchers.remove(*[Watcher.from_addr(addr) for addr in options['rm']])



# This file is part of django-ca (https://github.com/mathiasertl/django-ca).
#
# django-ca is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# django-ca is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with django-ca.  If not,
# see <http://www.gnu.org/licenses/>.

from cryptography import x509

from django import template
from django.contrib.admin.templatetags.admin_modify import submit_row

from ..utils import add_colons
from ..utils import bytes_to_hex
from ..utils import format_general_name
from ..utils import format_name
from ..utils import format_relative_name
from ..utils import int_to_hex

register = template.Library()


register.filter('format_name', format_name)
register.filter('format_relative_name', format_relative_name)


@register.filter
def format_general_names(value):
    """A template tag to format general names.

    Note that currently general names always occur as list.
    """
    return [format_general_name(v) for v in value]


@register.filter
def as_hex(value):
    """Takes a bytes value and returns its hex representation."""

    if isinstance(value, int):
        return add_colons(int_to_hex(value))
    else:
        return bytes_to_hex(value)


@register.filter
def oid_name(value):
    return value._name


@register.filter
def is_user_notice(value):
    return isinstance(value, x509.UserNotice)


@register.inclusion_tag('django_ca/admin/submit_line.html', takes_context=True)
def django_ca_certificate_submit_row(context):
    return submit_row(context)



# This file is part of django-ca (https://github.com/mathiasertl/django-ca).
#
# django-ca is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# django-ca is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with django-ca.  If not,
# see <http://www.gnu.org/licenses/>.


class RemovedInDjangoCA18Warning(DeprecationWarning):
    """Warning if a feature will be removed in the next django-ca version."""
    pass


class RemovedInDjangoCA19Warning(PendingDeprecationWarning):
    """Warning if a feature will be removed in the version after the next django-ca version."""
    pass


RemovedInNextVersionWarning = RemovedInDjangoCA18Warning



# This file is part of django-ca (https://github.com/mathiasertl/django-ca).
#
# django-ca is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# django-ca is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with django-ca.  If not,
# see <http://www.gnu.org/licenses/>.

"""Default Django app configuration.

.. seealso:: https://docs.djangoproject.com/en/dev/ref/applications/
"""

from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _


class DjangoCAConfig(AppConfig):  # pylint: disable=missing-class-docstring
    name = 'django_ca'
    verbose_name = _('Certificate Authority')



#!/usr/bin/env python3
#
# This file is part of django-ca (https://github.com/mathiasertl/django-ca).
#
# django-ca is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# django-ca is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with django-ca.  If not,
# see <http://www.gnu.org/licenses/>.

"""setuptools based setup.py file for django-ca."""

import os
import sys

from setuptools import find_packages
from setuptools import setup

LONG_DESCRIPTION = """django-ca is a tool to manage TLS certificate authorities and easily issue and revoke
certificates. It is based `cryptography <https://cryptography.io/>`_ and `Django
<https://www.djangoproject.com/>`_. It can be used as an app in an existing Django project or stand-alone with
the basic project included.  Everything can be managed via the command line via `manage.py` commands - so no
webserver is needed, if you’re happy with the command-line.

Features:

* Set up a secure local certificate authority in just a few minutes.
* Written in Python Python3.6+, requires Django 2.2+ or later.
* Manage your entire certificate authority from the command line and/or via Djangos admin interface.
* Get email notifications about certificates about to expire.
* Certificate validation using Certificate Revocation Lists (CRLs) and via an included OCSP responder.

Please see https://django-ca.readthedocs.org for more extensive documentation.
"""

install_requires = [
    'django>=2.2',
    'asn1crypto>=1.0.1',
    'cryptography>=2.8',
    'django-object-actions>=1.1',
    'idna>=2.9',
    'packaging',
]

package_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'ca')
package_root = os.path.join(package_path, 'django_ca')

if os.path.exists(package_path):
    sys.path.insert(0, package_path)

# https://packaging.python.org/guides/single-sourcing-package-version/
import django_ca  # NOQA: E402, pylint: disable=wrong-import-position


def find_package_data(path):
    """Find static package data for given path."""
    data = []
    prefix = len(package_root) + 1
    for root, _dirs, files in os.walk(os.path.join(package_root, path)):
        for file in files:
            data.append(os.path.join(root, file)[prefix:])
    return data


package_data = find_package_data('static') + find_package_data('templates')

setup(
    name='django-ca',
    version=django_ca.__version__,
    description='A Django app providing a SSL/TLS certificate authority.',
    long_description=LONG_DESCRIPTION,
    author='Mathias Ertl',
    author_email='mati@er.tl',
    url='https://github.com/mathiasertl/django-ca',
    packages=find_packages('ca', exclude=('ca', 'django_ca.tests')),
    package_dir={'': 'ca'},
    package_data={'': package_data},
    python_requires='>=3.6',
    zip_safe=False,  # because of the static files
    install_requires=install_requires,
    extras_require={
        'acme': ['acme>=1.10', 'josepy>=1.3.0', 'requests'],
        'redis': ['hiredis>=1.0', 'redis>=3.2', 'django-redis-cache>=1.8.0'],
        'celery': ['celery>=4.3'],
    },
    classifiers=[
        'Development Status :: 4 - Beta',
        'Framework :: Django :: 2.2',
        'Framework :: Django :: 3.0',
        'Framework :: Django :: 3.1',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Security :: Cryptography',
        'Topic :: Security',
    ],
)



# This file is part of django-ca (https://github.com/mathiasertl/django-ca).
#
# django-ca is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# django-ca is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with django-ca.  If not,
# see <http://www.gnu.org/licenses/>.

"""
**django-ca** adds a few custom Django signals to important events to let you execute custom actions when
these events happen. Please see `Djangos documentation on signals
<https://docs.djangoproject.com/en/dev/ref/signals/>`_ for further information on how to use signals.

If you use **django-ca** as :ref:`standalone project <as-standalone>`, use the :ref:`CA_CUSTOM_APPS
<settings-ca-custom-apps>` setting to add a custom django app. Please see the `Django documentation on apps
<https://docs.djangoproject.com/en/dev/ref/applications/>`_ if you need help on writing Django apps.
"""

import django.dispatch

pre_create_ca = django.dispatch.Signal()
"""Called before a new certificate authority is created.

Parameters
----------

name : str
    The name of the future CA.
**kwargs
"""

post_create_ca = django.dispatch.Signal()
"""Called after a new certificate authority was created.

Parameters
----------

ca : :py:class:`~django_ca.models.CertificateAuthority`
    The certificate authority that was just created.
"""


pre_issue_cert = django.dispatch.Signal()
"""Called before a new certificate is issued.

Parameters
----------

ca
csr
**kwargs
    All additional parameters passed to :py:meth:`~django_ca.profiles.Profile.create_cert`, but normalized to
    expected values. For example, this signal will receive a :py:class:`~django_ca.extensions.KeyUsage`
    instance, regardless of how the parameter was actually passed.
"""

post_issue_cert = django.dispatch.Signal()
"""Called after a new certificate was issued.

Parameters
----------

cert : :py:class:`~django_ca.models.Certificate`
    The certificate that was just issued.
"""

pre_revoke_cert = django.dispatch.Signal()
"""Called before a certificate is revoked.

Parameters
----------

ca
csr
**kwargs
"""

post_revoke_cert = django.dispatch.Signal()
"""Called after a certificate was revoked

Parameters
----------

cert : :py:class:`~django_ca.models.Certificate`
    The certificate that was just revoked.
"""



# This file is part of django-ca (https://github.com/mathiasertl/django-ca).
#
# django-ca is free software: you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# django-ca is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with django-ca.  If not,
# see <http://www.gnu.org/licenses/>.

"""Management command to import a certificate.

.. seealso:: https://docs.djangoproject.com/en/dev/howto/custom-management-commands/
"""

import argparse

from cryptography import x509
from cryptography.hazmat.backends import default_backend

from django.core.management.base import CommandError

from ...models import Certificate
from ..base import BaseCommand


class Command(BaseCommand):  # pylint: disable=missing-class-docstring
    help = """Import an existing certificate.

The authority that that signed the certificate must exist in the database."""

    def add_arguments(self, parser):
        self.add_ca(parser, allow_disabled=False)
        parser.add_argument('pub', help='Path to the public key (PEM or DER format).',
                            type=argparse.FileType('rb'))

    def handle(self, pub, **options):  # pylint: disable=arguments-differ
        pub_data = pub.read()

        # close reader objects (otherwise we get a ResourceWarning)
        pub.close()

        # load public key
        try:
            pub_loaded = x509.load_pem_x509_certificate(pub_data, default_backend())
        except Exception:  # pylint: disable=broad-except
            try:
                pub_loaded = x509.load_der_x509_certificate(pub_data, default_backend())
            except Exception as ex:
                raise CommandError('Unable to load public key.') from ex

        cert = Certificate(ca=options['ca'])
        cert.x509 = pub_loaded
        cert.save()
